# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import json
import os.path
from pathlib import Path
import shutil
from typing import Dict, List
from discopop_library.CodeGenerator.CodeGenerator import from_json_strings
from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments
from discopop_library.PatchGenerator.diffs import get_diffs_from_modified_code
from discopop_library.PatchGenerator.from_optimizer_output import from_optimizer_output


def from_json_patterns(
    arguments: PatchGeneratorArguments,
    patterns_by_type: Dict[str, List[str]],
    file_mapping: Dict[int, Path],
    patch_generator_dir: str,
):
    # generate code modifications from each suggestion, create a patch and store the patch
    # using the suggestions unique id
    if arguments.verbose:
        print("Generating modified code...")
    for suggestion_type in patterns_by_type:
        if suggestion_type == "version":
            continue
        for suggestion in patterns_by_type[suggestion_type]:
            if suggestion_type == "optimizer_output":
                from_optimizer_output(file_mapping, patterns_by_type, suggestion, arguments, patch_generator_dir)
                continue
            if arguments.verbose:
                print("Suggestion: ", suggestion)
            file_id_to_modified_code: Dict[int, str] = from_json_strings(
                file_mapping,
                {suggestion_type: [suggestion]},
                CC=arguments.CC,
                CXX=arguments.CXX,
                skip_compilation_check=True,
            )
            # create patches from the modified codes
            file_id_to_patches: Dict[int, str] = get_diffs_from_modified_code(
                file_mapping, file_id_to_modified_code, arguments
            )
            if arguments.verbose:
                print("Patches: ", file_id_to_patches)
            # clear old results and save patches
            suggestion_dict = json.loads(suggestion)
            suggestion_id = suggestion_dict["pattern_id"]
            suggestion_folder_path = os.path.join(patch_generator_dir, str(suggestion_id))
            if arguments.verbose:
                print("Saving patches for suggestion: ", suggestion_id)
            if os.path.exists(suggestion_folder_path):
                shutil.rmtree(suggestion_folder_path)
            os.mkdir(suggestion_folder_path)
            for file_id in file_id_to_patches:
                patch_path = os.path.join(suggestion_folder_path, str(file_id) + ".patch")
                with open(patch_path, "w") as f:
                    f.write(file_id_to_patches[file_id])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast
from discopop_explorer.PEGraphX import LoopNode
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.Update import Update
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern


def remove_loop_index_updates(
    experiment: Experiment, best_configuration: OptimizerOutputPattern, arguments: OptimizerArguments
) -> OptimizerOutputPattern:
    to_be_removed: List[Update] = []
    for update in best_configuration.data_movement:
        # check for loop nodes as update targets
        if type(data_at(experiment.optimization_graph, update.target_node_id)) == Loop:
            # get loop indices from PEGraph
            loop_cu_id = data_at(experiment.optimization_graph, update.target_node_id).original_cu_id
            if loop_cu_id is None:
                continue
            loop_indices = cast(LoopNode, experiment.detection_result.pet.node_at(loop_cu_id)).loop_indices
            # check for loop indices as targeted varbiables
            if update.write_data_access.var_name in loop_indices:
                # found update to loop index
                if arguments.verbose:
                    print("ignoring update: ", str(update), " , since it targets a loop index.")
                to_be_removed.append(update)
    # remove identified loop index updates
    for tbr in to_be_removed:
        if tbr in best_configuration.data_movement:
            best_configuration.data_movement.remove(tbr)

    # ensure correct formatting
    if arguments.verbose:
        if len(to_be_removed) > 0:
            print()

    if arguments.verbose:
        print("Removed loop index updates")
        for update in best_configuration.data_movement:
            print("# ", update)
        print()

    return best_configuration



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import logging
import os
import re
import subprocess
from contextlib import suppress
from typing import List

from .utils import recursive_scandir, is_compile, is_link, get_library


class DiscopopCpp:
    def __init__(
        self,
        cugeneration: bool,
        dpinstrumentation: bool,
        dpreduction: bool,
        clang_path: str,
    ):
        self.cugeneration = cugeneration
        self.dpinstrumentation = dpinstrumentation
        self.dpreduction = dpreduction
        self.clang_path = clang_path

    def update_filemapping(self):
        cwd = os.getcwd()
        with suppress(FileNotFoundError):
            # Do not regenerate if FileMapping.txt is still up-to-date.
            filemapping_mtime = os.stat("FileMapping.txt").st_mtime
            if os.stat(cwd).st_mtime < filemapping_mtime and all(
                [not entry.is_dir() or entry.stat().st_mtime < filemapping_mtime for entry in recursive_scandir(cwd)]
            ):
                return
        logging.info("Generating FileMapping.txt.")
        with open("FileMapping.txt", "w") as fd:
            i = 1
            for entry in sorted(recursive_scandir(cwd), key=lambda e: e.path):
                if entry.is_file() and re.match(r"^[^.].+\.(?:c|cc|cpp|h|hpp|ipp)$", entry.name):
                    print(f"{i}\t{entry.path}", file=fd)
                    i += 1

    def wrap_clang_args(self, clang_args: List[str]) -> List[str]:
        args = [self.clang_path]
        if is_compile(clang_args):
            if any([self.cugeneration, self.dpinstrumentation, self.dpreduction]):
                self.update_filemapping()
                args += ["-g", "-O0", "-fno-discard-value-names"]
                if self.cugeneration:
                    # clang++ -g -O0 -fno-discard-value-names -Xclang -load \
                    #   -Xclang ${DISCOPOP_INSTALL}/libi/LLVMCUGeneration.so \
                    #   -mllvm -fm-path -mllvm ./FileMapping.txt -c <C_File>
                    args += ["-Xclang", "-load", "-Xclang", get_library("LLVMCUGeneration.so")]
                if self.dpinstrumentation:
                    # clang++ -g -O0 -fno-discard-value-names -Xclang -load \
                    #   -Xclang ${DISCOPOP_INSTALL}/libi/LLVMDPInstrumentation.so \
                    #   -mllvm -fm-path -mllvm ./FileMapping.txt -c <C_File> -o out.o
                    args += ["-Xclang", "-load", "-Xclang", get_library("LLVMDPInstrumentation.so")]
                if self.dpreduction:
                    # clang++ -g -O0 -fno-discard-value-names -Xclang -load \
                    #   -Xclang ${DISCOPOP_INSTALL}/libi/LLVMDPReduction.so \
                    #   -mllvm -fm-path -mllvm ./FileMapping.txt -c <C_File> -o out.o
                    args += ["-Xclang", "-load", "-Xclang", get_library("LLVMDPReduction.so")]
                args += ["-mllvm", "-fm-path", "-mllvm", "./FileMapping.txt"]
        args += clang_args
        if is_link(clang_args):
            if self.dpinstrumentation or self.dpreduction:
                # clang++ out.o -L${DISCOPOP_INSTALL}/rtlib -lDiscoPoP_RT -lpthread
                args += [
                    f"-L{os.path.dirname(get_library('libDiscoPoP_RT.a'))}",
                    "-lDiscoPoP_RT",
                    "-lpthread",
                ]
        return args

    def invoke(self, clang_args: List[str]) -> subprocess.CompletedProcess:  # type: ignore
        args = self.wrap_clang_args(clang_args)
        logging.info(" ".join(args))
        return subprocess.run(args)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
from typing import List, cast

from discopop_library.PatchApplicator.PatchApplicatorArguments import PatchApplicatorArguments


def list_applied_suggestions(arguments: PatchApplicatorArguments, applied_suggestions_file: str) -> List[str]:
    with open(applied_suggestions_file, "r") as f:
        applied_suggestions = json.loads(f.read())
        return cast(List[str], applied_suggestions["applied"])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, Set, Optional, cast, Dict, Tuple

from alive_progress import alive_bar  # type: ignore

from discopop_explorer.PEGraphX import (
    PEGraphX,
    CUNode,
    NodeType,
    EdgeType,
    DepType,
    NodeID,
    LoopNode,
    Node,
)
from discopop_explorer.variable import Variable
from .GPULoop import GPULoopPattern
from .GPUMemory import map_node
from ..PatternInfo import PatternInfo


class GPURegionInfo(PatternInfo):
    """Class, that represents an identified GPU Region"""

    contained_loops: List[GPULoopPattern]
    contained_cu_ids: List[NodeID]
    map_from_vars: List[str]
    map_to_vars: List[str]
    map_to_from_vars: List[str]
    map_alloc_vars: List[str]
    map_delete_vars: List[str]
    produced_vars: List[str]
    consumed_vars: List[str]
    project_folder_path: str

    def __init__(
        self,
        pet: PEGraphX,
        contained_loops: List[GPULoopPattern],
        contained_cu_ids: List[NodeID],
        map_to_vars: List[str],
        map_from_vars: List[str],
        map_to_from_vars: List[str],
        map_alloc_vars: List[str],
        map_delete_vars: List[str],
        produced_vars: List[str],
        consumed_vars: List[str],
        project_folder_path: str,
    ):
        node_id = sorted([loop.nodeID for loop in contained_loops])[0]
        PatternInfo.__init__(self, pet.node_at(node_id))
        self.contained_loops = contained_loops
        self.contained_cu_ids = contained_cu_ids
        self.map_to_vars = map_to_vars
        self.map_from_vars = map_from_vars
        self.map_to_from_vars = map_to_from_vars
        self.map_alloc_vars = map_alloc_vars
        self.map_delete_vars = map_delete_vars
        self.produced_vars = produced_vars
        self.consumed_vars = consumed_vars
        self.start_line = min([l.start_line for l in contained_loops])
        self.end_line = max([l.end_line for l in contained_loops])
        self.project_folder_path = project_folder_path

    def __str__(self):
        raise NotImplementedError()  # used to identify necessity to call to_string() instead

    def to_string(self, pet: PEGraphX):
        contained_loops_str = "\n" if len(self.contained_loops) > 0 else ""
        for loop in self.contained_loops:
            loop_str = loop.to_string(pet)
            # pretty printing
            loop_str = "".join(["\t" + s + "\n" for s in loop_str.split("\n")])
            contained_loops_str += loop_str

        return (
            f"GPU Region at: {self.node_id}\n"
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            f"Map to: {self.map_to_vars}\n"
            f"Map from: {self.map_from_vars}\n"
            f"Map to/from: {self.map_to_from_vars}\n"
            f"Map alloc: {self.map_alloc_vars}\n"
            f"Map delete: {self.map_delete_vars}\n"
            f"Contained patterns: {contained_loops_str}\n"
        )

    def get_entry_cus(self, pet: PEGraphX) -> List[str]:
        """returns a list of contained cus with predecessors outside the GPU Region,
        i.e. entry points into the region"""
        entry_cus: List[str] = []
        for contained_cu_id in self.contained_cu_ids:
            # check if contained_cu has only predecessors inside the gpu region
            for predecessor_id, _, _ in pet.in_edges(contained_cu_id, EdgeType.SUCCESSOR):
                if predecessor_id not in self.contained_cu_ids:
                    # found entry node
                    entry_cus.append(contained_cu_id)
        # remove duplicates
        entry_cus = list(set(entry_cus))
        return entry_cus

    def get_exit_cus(self, pet: PEGraphX) -> Tuple[List[str], List[str]]:
        """returns a list of contained cus with successors outside the GPU Region,
        i.e. exit points of the region. as well as a list of their successors, i.e. the nodes after the region
        """
        contained_exit_cus: List[str] = []
        outside_cus: List[str] = []
        for contained_cu_id in self.contained_cu_ids:
            # check if contained_cu_id has only successors inside the gpu region
            for _, successor_id, _ in pet.out_edges(contained_cu_id, EdgeType.SUCCESSOR):
                if successor_id not in self.contained_cu_ids:
                    # found exit node
                    contained_exit_cus.append(contained_cu_id)
                    outside_cus.append(successor_id)
        # remove duplicates
        contained_exit_cus = list(set(contained_exit_cus))
        outside_cus = list(set(outside_cus))
        return contained_exit_cus, outside_cus


class GPURegions:
    cascadingLoopsInRegions: List[List[NodeID]]
    gpu_loop_patterns: List[GPULoopPattern]
    loopsInRegion: List[str]
    pet: PEGraphX
    numRegions: int
    cu_ids_by_region: Dict[Tuple[NodeID, ...], List[NodeID]]
    map_type_from_by_region: Dict[Tuple[str, ...], List[str]]
    map_type_to_by_region: Dict[Tuple[str, ...], List[str]]
    map_type_tofrom_by_region: Dict[Tuple[str, ...], List[str]]
    map_type_alloc_by_region: Dict[Tuple[str, ...], List[str]]
    map_type_delete_by_region: Dict[Tuple[str, ...], List[str]]
    produced_vars: Dict[Tuple[str, ...], List[str]]
    consumed_vars: Dict[Tuple[str, ...], List[str]]
    project_folder_path: str

    def __init__(self, pet, gpu_patterns, project_folder_path):
        self.loopsInRegion = []
        self.gpu_loop_patterns = gpu_patterns
        self.project_folder_path = project_folder_path
        self.cascadingLoopsInRegions = [[]]
        self.numRegions = 0
        self.pet = pet
        self.cu_ids_by_region = dict()
        self.map_type_from_by_region = dict()
        self.map_type_to_by_region = dict()
        self.map_type_tofrom_by_region = dict()
        self.map_type_alloc_by_region = dict()
        self.map_type_delete_by_region = dict()
        self.produced_vars = dict()  # includes implicitly mapped variables as well
        self.consumed_vars = dict()  # includes implicitly mapped variables as well

    def findGPULoop(self, nodeID: str) -> Optional[GPULoopPattern]:
        """

        :param nodeID:
        :return:
        """
        for i in self.gpu_loop_patterns:
            if i.node_id == nodeID:
                return i
        return None

    def reachableCUs(self, cuID: NodeID, nextCUID: NodeID) -> bool:
        """

        :param cuID:
        :param nextCUID:
        :return:
        """
        loop: LoopNode = cast(LoopNode, map_node(self.pet, cuID))
        nextLoop: LoopNode = cast(LoopNode, map_node(self.pet, nextCUID))

        # next loop is not a loop!
        if nextLoop.type != 2:
            return False

        loopFirstChild: Node = self.pet.direct_children(loop)[0]
        CUIDsofLoop = self.pet.out_edges(loopFirstChild.id, EdgeType.SUCCESSOR)
        lastCUofLoop = map_node(self.pet, CUIDsofLoop[-1][1])
        nextLoopFirstChild: Node = self.pet.direct_children(nextLoop)[0]
        successors = self.pet.out_edges(lastCUofLoop.id, EdgeType.SUCCESSOR)
        if successors:
            if nextLoopFirstChild.id == successors[0][1]:
                return True
            else:
                return False
        return False

    def identifyGPURegions(self) -> None:
        """

        :return:
        """
        regionNum = 0
        # find the next loop of each loop in the source code
        self.findNextLoop()

        self.cascadingLoopsInRegions[regionNum].append(self.gpu_loop_patterns[0].nodeID)
        print("\tidentify cascading loops...")
        with alive_bar(len(self.gpu_loop_patterns)) as progress_bar:
            for i in range(0, len(self.gpu_loop_patterns)):
                if self.gpu_loop_patterns[i].nextLoop is not None:
                    if map_node(self.pet, cast(NodeID, self.gpu_loop_patterns[i].nextLoop)).type == 2:
                        if self.reachableCUs(
                            self.gpu_loop_patterns[i].nodeID,
                            cast(NodeID, self.gpu_loop_patterns[i].nextLoop),
                        ):
                            self.cascadingLoopsInRegions[regionNum].append(
                                cast(NodeID, self.gpu_loop_patterns[i].nextLoop)
                            )
                        else:
                            regionNum += 1
                            self.cascadingLoopsInRegions.append([])
                            self.cascadingLoopsInRegions[regionNum].append(
                                cast(NodeID, self.gpu_loop_patterns[i].nextLoop)
                            )
                progress_bar()

        self.numRegions = regionNum + 1

    def determineDataMapping(self) -> None:
        """determine data mapping for each GPU Region and their contained GPU Loops."""

        print("\tdetermine data mapping...")
        with alive_bar(len(self.cascadingLoopsInRegions)) as progress_bar:
            for region in self.cascadingLoopsInRegions:
                # determine CUs which belong to the region (= which are located inside the region)
                region_cus: List[Node] = []
                region_loop_patterns: List[GPULoopPattern] = []
                for loop_id in region:
                    loop_node: LoopNode = cast(LoopNode, self.pet.node_at(loop_id))
                    gpu_lp: GPULoopPattern = [p for p in self.gpu_loop_patterns if p.parentLoop == loop_id][0]
                    region_loop_patterns.append(gpu_lp)
                    region_cus += [cu for cu in self.pet.subtree_of_type(loop_node) if cu not in region_cus]
                    # add loop initialization to region cus (predecessor of first child of loop, if positions are suitable)
                    loop_entry_cu = self.pet.out_edges(loop_id, EdgeType.CHILD)[0][1]
                    predecessors = [s for s, t, d in self.pet.in_edges(loop_entry_cu, EdgeType.SUCCESSOR)]
                    for predecessor_id in predecessors:
                        predecessor_node = self.pet.node_at(predecessor_id)
                        if (
                            predecessor_node.start_line >= loop_node.start_line
                            and predecessor_node.end_line <= loop_node.end_line
                        ):
                            # predecessor is loop initialization
                            if predecessor_node not in region_cus:
                                region_cus.append(predecessor_node)

                self.cu_ids_by_region[tuple(region)] = [n.id for n in region_cus]

                # determine start and end line of region
                region_start_line = min([cu.start_line for cu in region_cus])
                region_end_line = max([cu.end_line for cu in region_cus])

                # determine variables which are written outside the region and read inside
                consumed_vars: List[str] = []
                for cu in region_cus:
                    in_dep_edges = self.pet.out_edges(cu.id, EdgeType.DATA)

                    # var is consumed, if incoming RAW dep exists
                    for sink_cu_id, source_cu_id, dep in in_dep_edges:
                        # unpack dep for sake of clarity
                        sink_line = dep.sink_line
                        source_line = dep.source_line
                        var_name = dep.var_name

                        if self.pet.node_at(source_cu_id) not in region_cus:
                            if dep.dtype == DepType.RAW:
                                if dep.var_name not in consumed_vars and dep.var_name is not None:
                                    consumed_vars.append(dep.var_name)

                # determine variables which are read afterwards and written in the region
                produced_vars: List[str] = []
                for cu in region_cus:
                    out_dep_edges = self.pet.in_edges(cu.id, EdgeType.DATA)
                    # var is produced, if outgoing RAW or WAW dep exists
                    for sink_cu_id, source_cu_id, dep in out_dep_edges:
                        # unpack dep for sake of clarity
                        sink_line = dep.sink_line
                        source_line = dep.source_line
                        var_name = dep.var_name
                        if self.pet.node_at(sink_cu_id) not in region_cus:
                            if dep.dtype in [DepType.RAW, DepType.WAW]:
                                if dep.var_name not in produced_vars and dep.var_name is not None:
                                    produced_vars.append(dep.var_name)

                # gather consumed, produced, allocated and deleted variables from mapping information
                map_to_vars: List[str] = []
                for loop_pattern in region_loop_patterns:
                    map_to_vars += [v.name for v in loop_pattern.map_type_to + loop_pattern.map_type_tofrom]
                map_to_vars = list(set(map_to_vars))

                map_from_vars: List[str] = []
                for loop_pattern in region_loop_patterns:
                    map_from_vars += [v.name for v in loop_pattern.map_type_from + loop_pattern.map_type_tofrom]
                map_from_vars = list(set(map_from_vars))

                map_alloc_vars: List[str] = []
                for loop_pattern in region_loop_patterns:
                    map_alloc_vars += [v.name for v in loop_pattern.map_type_alloc]
                map_alloc_vars = list(set(map_alloc_vars))

                map_to_from_vars = [var for var in map_to_vars if var in map_from_vars]

                # allocate unknown variables
                map_alloc_vars += [
                    var for var in map_from_vars if var not in consumed_vars + map_to_vars + map_alloc_vars
                ]
                map_alloc_vars = list(set(map_alloc_vars))

                # store results
                self.map_type_from_by_region[tuple(region)] = [
                    var for var in map_from_vars if var not in map_to_from_vars
                ]
                self.map_type_to_by_region[tuple(region)] = [var for var in map_to_vars if var not in map_to_from_vars]
                self.map_type_tofrom_by_region[tuple(region)] = map_to_from_vars
                self.map_type_alloc_by_region[tuple(region)] = map_alloc_vars
                self.map_type_delete_by_region[tuple(region)] = []
                self.produced_vars[tuple(region)] = produced_vars
                self.consumed_vars[tuple(region)] = consumed_vars

                progress_bar()

    def old_mapData(self) -> None:
        """
        :return:
        """

        for i in range(0, self.numRegions):
            for j in self.cascadingLoopsInRegions[i]:
                gpuLoop = self.findGPULoop(j)
                if gpuLoop is None:
                    continue
                gpuLoop.printGPULoop()
        print(f"==============================================")
        for i in range(0, self.numRegions):
            t: int = len(self.cascadingLoopsInRegions[i]) - 1
            firstNodeID = self.cascadingLoopsInRegions[i][0]
            lastNodeID = self.cascadingLoopsInRegions[i][t]
            fn = map_node(self.pet, firstNodeID)
            ln = map_node(self.pet, lastNodeID)
            start = fn.start_line
            end = ln.end_line
            gpuRegionLoop = GPULoopPattern(self.pet, firstNodeID, start, end, 1000, self.project_folder_path)
            visitedVars: Set[Variable] = set()
            while t >= 0:
                loopIter = self.findGPULoop(
                    self.cascadingLoopsInRegions[i][t]
                )  # tmp_result contains GPU loops inside the parent region

                if loopIter is None:
                    t -= 1
                    continue

                varis: Set[Variable] = set([])
                varis.update(loopIter.map_type_alloc)
                varis.update(loopIter.map_type_to)
                varis.update(loopIter.map_type_from)
                varis.update(loopIter.map_type_tofrom)
                # loopIter.printGPULoop()
                for v in varis:
                    if loopIter.findMappedVar("from", v):
                        if v in visitedVars:
                            if gpuRegionLoop.findMappedVar("to", v):
                                # // cout << "gpuRegionLoop to: " << get_var_name(v) << endl;
                                gpuRegionLoop.map_type_to.remove(v)
                                # gpuRegionLoop.map_type_to.erase(remove(gpuRegionLoop.map_type_to.begin(),
                                #                                       gpuRegionLoop.map_type_to.end(), v),
                                #                                gpuRegionLoop.map_type_to.end())
                                gpuRegionLoop.map_type_alloc.append(v)
                                visitedVars.add(v)
                            else:
                                gpuRegionLoop.map_type_from.append(v)
                                visitedVars.add(v)
                        else:
                            # // cout << "visited from: " << get_var_name(v) << endl;
                            gpuRegionLoop.map_type_from.append(v)
                            visitedVars.add(v)
                    elif loopIter.findMappedVar("to", v):
                        if v in visitedVars:
                            # // cout << "loopIter to: " << get_var_name(v) << endl;
                            if gpuRegionLoop.findMappedVar("alloc", v):
                                # // cout << "gpuRegionLoop alloc: " << get_var_name(v) << endl;
                                gpuRegionLoop.map_type_alloc.remove(v)

                                gpuRegionLoop.map_type_to.append(v)
                                visitedVars.add(v)
                        else:
                            # // cout << "visited to: " << get_var_name(v) << endl;
                            gpuRegionLoop.map_type_to.append(v)
                            visitedVars.add(v)
                    elif loopIter.findMappedVar("tofrom", v):
                        if v not in visitedVars:
                            # // cout << "loopIter tofrom: " << get_var_name(v) << endl;
                            gpuRegionLoop.map_type_tofrom.append(v)
                            visitedVars.add(v)
                        else:
                            if gpuRegionLoop.findMappedVar("to", v):
                                gpuRegionLoop.map_type_to.remove(v)
                                gpuRegionLoop.map_type_alloc.append(v)
                            elif gpuRegionLoop.findMappedVar("from", v):
                                gpuRegionLoop.map_type_from.remove(v)
                                gpuRegionLoop.map_type_tofrom.append(v)
                    elif loopIter.findMappedVar("alloc", v):
                        if v not in visitedVars:
                            gpuRegionLoop.map_type_alloc.append(v)
                            visitedVars.add(v)
                t -= 1

            # gpuRegionLoop.printGPULoop()  # only prints first loop of the region.
            # first loop of the region is used to store the map_to behavior
            # last loop if the region should be used to store the map_from behavior

    def findNextLoop(self) -> None:
        """

        :return:
        """

        i: int = 0
        while i < len(self.gpu_loop_patterns):
            skip: int = i
            for k in self.gpu_loop_patterns[skip].nestedLoops:
                """To find the next loop for a loop,
                we need to skip over the nested loops in the identified GPU loops.
                Also, some loops may not be do-all or reduction
                and not listed in the gpu_loop_patterns. We need to skip them
                Create a dummy GPULoopPattern and
                search for the gpu loops with the same NodeID
                """
                dd = GPULoopPattern(
                    self.pet,
                    k,
                    self.gpu_loop_patterns[skip].start_line,
                    self.gpu_loop_patterns[skip].end_line,
                    self.gpu_loop_patterns[skip].iteration_count,
                    self.project_folder_path,
                )
                for ii in self.gpu_loop_patterns:
                    if dd.nodeID == ii.nodeID:  # TODO: check remaining attributes too
                        i += 1

            if (i + 1) < len(self.gpu_loop_patterns):
                self.gpu_loop_patterns[skip].nextLoop = self.gpu_loop_patterns[i + 1].node_id

            i += 1

    def getParentLoop(self, node: CUNode) -> bool:
        """

        :param node:
        :return:
        """
        main: Node = self.pet.main
        path: List[Node] = self.pet.path(map_node(self.pet, main.id), node)
        path.reverse()

        for parent in path:
            if not parent.id == node.id and parent.type == NodeType.LOOP:
                return True
        return False

    def get_gpu_region_info(self, pet: PEGraphX, project_folder_path: str) -> List[GPURegionInfo]:
        """Construct GPURegionInfo representations of all identified GPU Regions and return them in a list"""
        gpu_region_info: List[GPURegionInfo] = []
        print("\tget gpu region info...")
        with alive_bar(len(self.cascadingLoopsInRegions)) as progress_bar:
            for region in self.cascadingLoopsInRegions:
                contained_loop_patterns = [pattern for pattern in self.gpu_loop_patterns if pattern.nodeID in region]
                # save OpenMP constructs in GPULoops for the exporting to JSON
                for loop in contained_loop_patterns:
                    loop.save_omp_constructs(pet, project_folder_path)
                device_cu_ids = self.cu_ids_by_region[tuple(region)]
                for loop in contained_loop_patterns:
                    for func_node_id in loop.called_functions:
                        for child in pet.direct_children(pet.node_at(func_node_id)):
                            device_cu_ids.append(child.id)

                device_cu_ids = list(set(device_cu_ids))
                current_info = GPURegionInfo(
                    self.pet,
                    contained_loop_patterns,
                    device_cu_ids,
                    self.map_type_to_by_region[tuple(region)],
                    self.map_type_from_by_region[tuple(region)],
                    self.map_type_tofrom_by_region[tuple(region)],
                    self.map_type_alloc_by_region[tuple(region)],
                    self.map_type_delete_by_region[tuple(region)],
                    self.produced_vars[tuple(region)],
                    self.consumed_vars[tuple(region)],
                    project_folder_path,
                )
                gpu_region_info.append(current_info)
                progress_bar()
        return gpu_region_info



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

import matplotlib  # type: ignore
import numpy as np

matplotlib.use("Agg")
from matplotlib import cm  # type: ignore
from matplotlib import pyplot as plt  # type: ignore
from matplotlib.ticker import LinearLocator  # type: ignore


class MicrobenchType(str, Enum):
    COPYIN = "COPYIN"
    COPYPRIVATE = "COPY_PRIVATE"
    DOALL = "DOALL"
    FIRSTPRIVATE = "FIRSTPRIVATE"
    SEPARATED = "SEPARATED"
    SHARED = "SHARED"
    PRIVATE = "PRIVATE"
    FOR = "FOR"


class MicrobenchDimension(str, Enum):
    REFERENCE = "Reference time in us"
    TEST = "Test time in us"
    OVERHEAD = "Overhead time in us"


class MicrobenchCoordinate(NamedTuple):
    threads: int
    workload: float
    iterations: int


class Microbench(ABC):
    @abstractmethod
    def getMeasurements(
        self,
    ) -> Dict[MicrobenchType, Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],]:
        raise TypeError("Pure virtual method called")

    @abstractmethod
    def toJSON(self) -> str:
        raise TypeError("Pure virtual method called")

    @abstractmethod
    def evaluateInterpolation(
        self,
        benchType: MicrobenchType,
        benchDim: MicrobenchDimension,
        benchCoord: Union[MicrobenchCoordinate, Tuple[int, float, float]],
    ) -> float:
        raise TypeError("Pure virtual method called")

    def plotInterpolation(
        self,
        type: MicrobenchType,
        dim: MicrobenchDimension,
        iterations: int,
        threads=range(1, 9),
        workloads=range(0, 201),
        file: Optional[Path] = None,
    ):
        logging.info("plotting interpolation to %s", str(file.absolute() if file else "<screen>"))
        # coords = list(self.getMeasurements()[type][dim].keys())
        # minWorkload = coords[0].workload
        # maxWorkload = coords[0].workload
        # for coord in coords[1:]:
        #    if coord.workload > maxWorkload:
        #        maxWorkload = coord.workload
        #    if coord.workload < minWorkload:
        #        minWorkload = coord.workload
        # threads = np.linspace(1, 8, 8, endpoint=True).tolist() # threads
        # workloads = np.linspace(minWorkload, maxWorkload, 50).tolist() # workload

        # x axis: threads
        # y axis: workloads
        # z axis: measurement values
        # (iterations are fixed based on input parameter)
        z = [self.evaluateInterpolation(type, dim, (i, j, iterations)) for i in threads for j in workloads]  # results
        X, Y = np.meshgrid(threads, workloads)
        Z = np.array(z).reshape(len(threads), len(workloads)).transpose()

        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})  # type: ignore
        # ignored type due to incompatibilities between local mypy and GitHub ci
        surf = ax.plot_surface(X, Y, Z, cmap=cm.coolwarm, linewidth=0, antialiased=False)  # type: ignore
        ax.set_zlim(0, np.nanmax(z))
        ax.zaxis.set_major_locator(LinearLocator(10))
        fig.colorbar(surf, shrink=0.5, aspect=5)

        plt.xlabel("Threads")
        plt.ylabel("Workload")
        plt.title(f"{type.value} {dim.value}:\ninterpolation for {iterations} iterations")
        if file is None:
            plt.show()
        else:
            plt.savefig(file)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import copy
import os
import re
import shlex
import subprocess
from pathlib import Path
from typing import List, Dict, Any, Optional

from discopop_library.CodeGenerator.classes.Enums import PragmaPosition
from discopop_library.CodeGenerator.classes.Line import Line
from discopop_library.CodeGenerator.classes.Pragma import Pragma


class ContentBuffer(object):
    lines: List[Line]
    max_line_num: int  # used to determine width of padding
    file_id: int
    next_free_region_id = 0
    line_type: Any
    compile_result_buffer: str

    def __init__(self, file_id: int, source_code_path: Path, tab_width: int = 4, line_type=Line):
        self.line_type = line_type
        self.file_id = file_id
        self.lines = []
        self.max_line_num = 0
        # load source code
        with open(source_code_path, "r") as f:
            lines = f.readlines()
            for idx, line in enumerate(lines):
                idx = idx + 1  # start with line number 1
                self.max_line_num = idx
                line_obj = self.line_type(idx, line_num=idx, content=line)
                self.lines.append(line_obj)
        self.compile_result_buffer = ""

    def print_lines(self):
        for line in self.lines:
            print(line)

    def __get_next_free_region_id(self) -> int:
        buffer = self.next_free_region_id
        self.next_free_region_id += 1
        return buffer

    def get_modified_source_code(self) -> str:
        result = ""
        for line in self.lines:
            if not line.content.endswith("\n"):
                line.content += "\n"
            result += line.content
        return result

    def append_line_before(self, parent_line_num: int, line: Line, match_indentation: bool = True):
        """Appends line before the specified parent_line_num"""
        for idx, potential_parent_line in enumerate(self.lines):
            if potential_parent_line.line_num == parent_line_num:
                if match_indentation:
                    line.content = potential_parent_line.get_indentation() + line.content
                self.lines.insert(idx, line)
                return

    def append_line_after(self, parent_line_num: int, line: Line, match_indentation: bool = True):
        """Appends line after the specified parent_line_num"""
        for idx, potential_parent_line in enumerate(self.lines):
            if potential_parent_line.line_num == parent_line_num:
                if idx + 1 < len(self.lines):
                    self.lines.insert(idx + 1, line)
                else:
                    self.lines.append(line)
                if match_indentation:
                    line.content = potential_parent_line.get_indentation() + line.content
                return

    def add_pragma(
        self,
        file_mapping: Dict[int, Path],
        pragma: Pragma,
        parent_regions: List[int],
        add_as_comment: bool = False,
        skip_compilation_check: bool = False,
        compile_check_command: Optional[str] = None,
        CC="clang",
        CXX="clang++",
        match_indentation: bool = True,
    ) -> bool:
        """insert pragma into the maintained list of source code lines.
        Returns True if the pragma resulted in a valid (resp. compilable) code transformation.
        Returns False if compilation of the modified code was not possible.
        In this case, no changes will be applied to the ContentBuffer Object."""
        if pragma.start_line is None or pragma.end_line is None:
            raise ValueError("Unsupported start or end line: ", pragma.start_line, "-", pragma.end_line)
        if pragma.file_id is None:
            raise ValueError("Unsupported file_id: ", pragma.file_id)
        if pragma.file_id != self.file_id:
            return True  # incorrect target file, ignore the pragma

        # create backup of ContentBuffer
        backup_lines = copy.deepcopy(self.lines)
        backup_max_line_num = self.max_line_num
        backup_file_id = self.file_id
        backup_next_free_region_id = self.next_free_region_id

        # construct line
        pragma_line = self.line_type(pragma.start_line)

        # adding as comment
        if add_as_comment:
            pragma_line.content = "//<DiscoPoP-IGNORED> "
        else:
            pragma_line.content = ""

        pragma_line.content += pragma.pragma_str
        pragma_line.belongs_to_regions = copy.deepcopy(parent_regions)
        # create new region if necessary
        if len(pragma.children) > 0:
            region_id = self.__get_next_free_region_id()
            pragma_line.owns_region = region_id
            pragma_line.belongs_to_regions.append(region_id)

        if pragma.pragma_position == PragmaPosition.BEFORE_START:
            self.append_line_before(pragma.start_line, pragma_line)
        elif pragma.pragma_position == PragmaPosition.AFTER_START:
            self.append_line_after(pragma.start_line, pragma_line)
        elif pragma.pragma_position == PragmaPosition.BEFORE_END:
            self.append_line_before(pragma.end_line, pragma_line)
        elif pragma.pragma_position == PragmaPosition.AFTER_END:
            self.append_line_after(pragma.end_line, pragma_line)
        else:
            raise ValueError("Unsupported pragma position: ", pragma.pragma_position)

        # update belonging information of contained lines
        tmp_start_line = (
            pragma.start_line if pragma.pragma_position == PragmaPosition.BEFORE_START else pragma.start_line + 1
        )
        tmp_end_line = pragma.end_line + 1
        for line_num in range(tmp_start_line, tmp_end_line):
            for line in self.lines:
                if line.line_num == line_num:
                    line.belongs_to_regions += [
                        n for n in pragma_line.belongs_to_regions if n not in line.belongs_to_regions
                    ]

        # append children to lines (mark as contained in region)
        for child_pragma in pragma.children:
            # set skip_compilation_check to true since compiling children pragmas on their own might not be successful.
            # As an example for that, '#pragma omp declare target' can be mentioned
            successful = self.add_pragma(
                file_mapping,
                child_pragma,
                pragma_line.belongs_to_regions,
                add_as_comment=add_as_comment,
                skip_compilation_check=True,
                compile_check_command=compile_check_command,
                CC=CC,
                CXX=CXX,
            )

            if not successful:
                print(self.compile_result_buffer)
                self.lines = backup_lines
                self.next_free_region_id = backup_next_free_region_id
                self.file_id = backup_file_id
                self.max_line_num = backup_max_line_num
                return False

        if skip_compilation_check:
            return True

        # check if the applied changes resulted in a compilable source code
        # create a temporary file to store the modified file contents
        if str(file_mapping[self.file_id]).endswith(".c"):
            compiler = CC
            tmp_file_name = str(file_mapping[self.file_id]) + ".discopop_tmp.c"
        else:
            compiler = CXX
            tmp_file_name = str(file_mapping[self.file_id]) + ".discopop_tmp.cpp"
        tmp_file_out_name = str(file_mapping[self.file_id]) + ".discopop_tmp.o"

        with open(tmp_file_name, "w+") as f:
            f.write(self.get_modified_source_code())
            f.flush()
            f.close()

        if compile_check_command is None:
            result = subprocess.run(
                [compiler, "-c", "-fopenmp", tmp_file_name, "-o", tmp_file_out_name],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        else:
            saved_dir = os.getcwd()
            # prepare environment variables if requested
            pattern = re.compile(r"""((?:[^\s\"\']|\"[^\"]*\"|'[^']*')+)""")
            splitted_compile_check_command = pattern.split(compile_check_command)[
                1::2
            ]  # compile_check_command.split(" ")
            print("SPLITTED: ", splitted_compile_check_command)
            to_be_removed = []
            for idx, elem in enumerate(splitted_compile_check_command):
                if "=" in elem and not elem.startswith("-"):
                    # set environment variable
                    var_name = elem[: elem.index("=")]
                    var_value = elem[elem.index("=") + 1 :]
                    print("SET ENVIRONMENT VAR: ")
                    print("\t", var_name)
                    print("\t\t", var_value)
                    # unpack var_value
                    if var_value.startswith('"') and var_value.endswith('"') and len(var_value) > 1:
                        var_value = var_value[1:-1]
                    os.environ[var_name] = var_value

                    to_be_removed.append(idx)
            for idx in sorted(to_be_removed, reverse=True):
                splitted_compile_check_command.pop(idx)

            #            print("COMPILE COMMAND: ", splitted_compile_check_command)
            #           result = subprocess.run(
            #               splitted_compile_check_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            #           )

            print("COMPILE COMMAND: ", shlex.split(compile_check_command))
            result = subprocess.run(shlex.split(compile_check_command), shell=True)

            os.chdir(saved_dir)
        compilation_successful = result.returncode == 0

        if os.path.exists(tmp_file_name):
            os.remove(tmp_file_name)
        if compilation_successful:
            if os.path.exists(tmp_file_out_name):
                os.remove(tmp_file_out_name)

        # if not, reset ContentBuffer to the backup and return False
        if not compilation_successful:
            self.lines = backup_lines
            self.next_free_region_id = backup_next_free_region_id
            self.file_id = backup_file_id
            self.max_line_num = backup_max_line_num
            self.compile_result_buffer += result.stdout.decode("utf-8") + "\n"
            self.compile_result_buffer += result.stderr.decode("utf-8") + "\n"
            self.compile_result_buffer += "==> Skipped pragma insertion due to potential compilation errors!\n"
            print(self.compile_result_buffer)
            return False
        return True



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Optional

from sympy import Symbol, Integer, Expr, Float  # type: ignore

from discopop_explorer.PEGraphX import NodeID
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload


class Loop(Workload):
    iterations: int
    position: str
    iterations_symbol: Symbol
    registered_child: Optional[CostModel]
    collapse_level: int

    def __init__(
        self,
        node_id: int,
        experiment: Experiment,
        cu_id: Optional[NodeID],
        discopop_workload: int,
        iterations: int,
        position: str,
        iterations_symbol: Optional[Symbol] = None,
        collapse_level: int = 1,
    ):
        self.position = position
        self.iterations = max(iterations, 1)  # to prevent dividing by 0 in case the loop has not been executed
        self.collapse_level = collapse_level

        if iterations_symbol is None:
            self.iterations_symbol = Symbol("loop_" + str(node_id) + "_pos_" + str(self.position) + "_iterations")
        else:
            self.iterations_symbol = iterations_symbol

        # create parallelizable_workload_symbol
        self.per_iteration_parallelizable_workload = Symbol(
            "loop_" + str(node_id) + "_pos_" + str(self.position) + "_per_iteration_parallelizable_workload"
        )
        self.per_iteration_sequential_workload = Symbol(
            "loop_" + str(node_id) + "_pos_" + str(self.position) + "_per_iteration_sequential_workload"
        )

        # calculate workload per iteration
        per_iteration_parallelizable_workload = discopop_workload / iterations

        super().__init__(
            node_id,
            experiment,
            cu_id,
            sequential_workload=self.per_iteration_sequential_workload,
            parallelizable_workload=self.per_iteration_parallelizable_workload,
            #            int(
            #                # per_iteration_parallelizable_workload
            #            ),  # todo this might be wrong! should be 1 instead, since children are registered individually
        )

        # register iteration symbol in environment
        experiment.register_free_symbol(self.iterations_symbol, value_suggestion=Integer(self.iterations))
        # register per iteration parallelizable workload symbol in environment
        experiment.register_free_symbol(
            self.per_iteration_parallelizable_workload,
            value_suggestion=Float(per_iteration_parallelizable_workload),
        )
        experiment.substitutions[self.per_iteration_sequential_workload] = Integer(0)
        self.registered_child = None

    # todo: note: it might be more beneficial to use the iterations "per entry" instead of the total amount of iterations
    # example:
    # for(100)      --> use 100
    #   for(100)    --> use 100 instead of 10000

    def get_hover_text(self) -> str:
        return (
            "WL: " + str(self.sequential_workload) + "\n" + "IT: " + str(self.iterations) + "\n"
            "Read: " + str([str(e) for e in self.read_memory_regions]) + "\n"
            "Write: " + str([str(e) for e in self.written_memory_regions]) + "\n"
            "Suggestion: " + str(self.suggestion) + "\n"
            "Collapse: " + str(self.collapse_level) + "\n"
            "Device ID: " + str(self.device_id)
        )

    def get_cost_model(self, experiment, all_function_nodes, current_device) -> CostModel:
        """Performance model of a workload consists of the workload itself.
        Individual Workloads are assumed to be not parallelizable.
        Workloads of Loop etc. are parallelizable."""

        # loop costs = self.sequential + overhead + iterations * per_iteration_workload * cost_modifier

        cm = CostModel(
            self.iterations_symbol
            * self.per_iteration_parallelizable_workload
            * self.cost_multiplier.parallelizable_costs,
            self.sequential_workload
            + self.overhead.sequential_costs
            + self.per_iteration_sequential_workload * self.iterations_symbol,
        )

        # substitute Expr(0) with Integer(0)
        cm.parallelizable_costs = cm.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})
        cm.sequential_costs = cm.sequential_costs.subs({Expr(Integer(0)): Integer(0)})

        return cm

    def register_child(self, other, experiment, all_function_nodes, current_device):
        """Registers a child node for the given model.
        Does not modify the stored model in self or other."""

        # at every time, only a single child is possible for each loop node
        self.registered_child = other
        experiment.substitutions[self.per_iteration_parallelizable_workload] = other.parallelizable_costs
        experiment.substitutions[self.per_iteration_sequential_workload] = other.sequential_costs

        cm = self.get_cost_model(experiment, all_function_nodes, current_device)
        cm.path_decisions += other.path_decisions
        return cm

    def register_successor(self, other):
        """Registers a successor node for the given model.
        Does not modify the stored model in self or other."""
        # sequential composition is depicted by simply adding the performance models
        return self.performance_model.parallelizable_plus_combine(other)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
import tkinter as tk
from tkinter import ttk
from typing import List

from DEPRECATED_discopop_wizard.classes.ExecutionConfiguration import ExecutionConfiguration
from DEPRECATED_discopop_wizard.screens.settings import show_settings_screen
from DEPRECATED_discopop_wizard.utils import support_scrolling


class MainScreen(object):
    configuration_frame: tk.Frame
    notebook: ttk.Notebook
    details_frame: tk.Frame
    results_frame: tk.Frame

    toggle_var_code_preview_show_metadata: tk.IntVar

    def __init__(self, wizard, window_frame: tk.Frame):
        self.wizard = wizard
        self.push_main_screen(wizard, window_frame)

    def push_main_screen(self, wizard, window_frame: tk.Frame):
        frame = tk.Frame(window_frame)
        frame.grid(row=1, column=1, sticky="nsew")

        # create horizontally split frames (configurations + details frame)
        horizontal_paned_window = ttk.PanedWindow(frame, orient="horizontal")
        horizontal_paned_window.pack(fill=tk.BOTH, expand=True)
        self.configuration_frame = tk.Frame(horizontal_paned_window)
        horizontal_paned_window.add(self.configuration_frame)
        self.notebook = ttk.Notebook(horizontal_paned_window)
        horizontal_paned_window.add(self.notebook, weight=1)

        self.details_frame = tk.Frame(self.notebook)
        self.results_frame = tk.Frame(self.notebook)
        self.optimizer_frame = tk.Frame(self.notebook)
        self.notebook.add(self.details_frame, text="Details")
        self.notebook.add(self.results_frame, text="Results")
        self.notebook.add(self.optimizer_frame, text="Optimizer (experimental)")

        self.build_configurations_frame(wizard)

        self.__build_menu_bar(wizard)

    def __build_menu_bar(self, wizard):
        # build menu bar
        optionsmenu = tk.Menu(wizard.menubar)
        wizard.menubar.add_cascade(label="Options", menu=optionsmenu)
        optionsmenu.add_command(label="Settings", command=lambda: show_settings_screen(wizard))
        # build code preview menu
        code_preview_menu = tk.Menu(optionsmenu)
        optionsmenu.add_cascade(label="Code Preview", menu=code_preview_menu)
        code_preview_menu.add_checkbutton(
            label="Show Metadata Regions",
            variable=wizard.tk_var_storage.toggle_var_code_preview_show_metadata_regions,
            onvalue=1,
            offvalue=0,
            command=lambda: wizard.tk_var_storage.toggle_code_preview_setting_action(),
        )
        code_preview_menu.add_checkbutton(
            label="Show Line Numbers",
            variable=wizard.tk_var_storage.toggle_var_code_preview_show_line_numbers,
            onvalue=1,
            offvalue=0,
            command=lambda: wizard.tk_var_storage.toggle_code_preview_setting_action(),
        )
        code_preview_menu.add_checkbutton(
            label="Show Live Variables",
            variable=wizard.tk_var_storage.toggle_var_code_preview_show_metadata_live_device_variables,
            onvalue=1,
            offvalue=0,
            command=lambda: wizard.tk_var_storage.toggle_code_preview_setting_action(),
        )
        code_preview_menu.add_checkbutton(
            label="Disable compile validation of inserted pragmas",
            variable=wizard.tk_var_storage.toggle_var_code_preview_disable_compile_check,
            onvalue=1,
            offvalue=0,
            command=lambda: wizard.tk_var_storage.toggle_code_preview_setting_action(),
        )

    def build_configurations_frame(self, wizard):
        # clear previous contents if existent
        for c in self.configuration_frame.winfo_children():
            c.destroy()
        # build configuration frame
        self.__display_execution_configurations(wizard)

    def __create_new_execution_configuration(self, wizard):
        execution_config = ExecutionConfiguration(wizard)
        execution_config.show_details_screen(wizard, self)

    def __display_execution_configurations(self, wizard):
        # based on https://blog.teclado.com/tkinter-scrollable-frames/
        # load configuration options
        configs: List[ExecutionConfiguration] = self.load_execution_configurations(wizard.config_dir)
        frame = tk.Frame(self.configuration_frame)
        frame.pack(fill=tk.BOTH, expand=True)
        tk.Label(frame, text="Configurations", font=wizard.style_font_bold, pady=10).pack()
        # add New.. Button
        tk.Button(frame, text="New..", command=lambda: self.__create_new_execution_configuration(wizard)).pack()

        tmp_frame = tk.Frame(frame)
        tmp_frame.pack(fill=tk.BOTH, expand=True)

        # create scrollable list of suggestions
        canvas = tk.Canvas(tmp_frame)
        scrollbar = tk.Scrollbar(tmp_frame, orient="vertical", command=canvas.yview)
        scrollable_frame = tk.Frame(canvas)
        scrollable_frame.bind("<Configure>", lambda e: canvas.configure(scrollregion=canvas.bbox("all")))
        canvas.create_window((0, 0), window=scrollable_frame, anchor="nw")
        canvas.configure(yscrollcommand=scrollbar.set)

        all_buttons: List[tk.Button] = []  # used to manage highlights when a different configuration is selected
        for row, config in enumerate(configs):
            button = config.get_as_button(wizard, self, scrollable_frame, all_buttons)
            button.pack(fill=tk.BOTH, expand=True)
            all_buttons.append(button)

        # support scrolling on all platforms (linux, windows, macOS)
        support_scrolling(canvas)

        canvas.pack(side="left", fill="both", expand=True)
        scrollbar.pack(side="right", fill="y")

    def load_execution_configurations(self, config_dir: str) -> List[ExecutionConfiguration]:
        execution_configs: List[ExecutionConfiguration] = []
        for filename in os.listdir(os.path.join(config_dir, "execution_configurations")):
            if not filename.endswith(".json"):
                continue
            with open(os.path.join(os.path.join(config_dir, "execution_configurations"), filename), "r") as json_file:
                config = ExecutionConfiguration(self.wizard)
                config.init_from_json(json_file)
                execution_configs.append(config)
        return execution_configs



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter as tk
from tkinter import ttk


class Console(object):
    parent_frame: tk.Frame
    progress_bar: ttk.Progressbar
    log_screen: tk.Text

    def __init__(self, parent_frame: tk.Frame):
        self.parent_frame = tk.Frame(parent_frame, bg="red")
        self.parent_frame.grid(row=1, column=1, sticky="nsew")
        self.__show_console()

    def print(self, msg: str):
        if not msg.endswith("\n"):
            msg = msg + "\n"
        self.log_screen.config(state=tk.NORMAL)
        self.log_screen.insert(tk.END, msg)
        self.log_screen.see(tk.END)
        self.log_screen.update()
        self.log_screen.config(state=tk.DISABLED)

    def clear(self):
        raise NotImplementedError("TODO")

    def start_progress(self):
        self.progress_bar.start(20)
        self.progress_bar.update()

    def stop_progress(self):
        self.progress_bar.stop()
        self.progress_bar.update()

    def __show_console(self):
        for c in self.parent_frame.winfo_children():
            c.destroy()
        # configure parent_frame
        self.parent_frame.rowconfigure(0, weight=1)
        self.parent_frame.columnconfigure(0, weight=1)
        # self.parent_frame.rowconfigure(2, weight=1)

        # create content frame and scroll bars
        self.log_screen = tk.Text(self.parent_frame, wrap=tk.NONE, height=8)
        self.log_screen.grid(row=0, column=0, sticky="nsew")
        self.log_screen.config(state=tk.DISABLED)

        # create a Scrollbar and associate it with the content frame
        y_scrollbar = ttk.Scrollbar(self.parent_frame, command=self.log_screen.yview)
        y_scrollbar.grid(row=0, column=1, sticky="nsew")
        self.log_screen["yscrollcommand"] = y_scrollbar.set
        x_scrollbar = ttk.Scrollbar(self.parent_frame, orient="horizontal", command=self.log_screen.xview)
        x_scrollbar.grid(row=1, column=0, columnspan=2, sticky="nsew")
        self.log_screen["xscrollcommand"] = x_scrollbar.set

        # create progress bar
        # self.progress_bar = ttk.Progressbar(self.parent_frame, orient=tk.HORIZONTAL, mode="determinate")
        # self.progress_bar.grid(row=2, column=0, columnspan=2, sticky="nsew")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os

from discopop_library.PatchApplicator.PatchApplicatorArguments import PatchApplicatorArguments
from discopop_library.PatchApplicator.apply import apply_patches
from discopop_library.PatchApplicator.clear import clear_patches
from discopop_library.PatchApplicator.list import list_applied_suggestions
from discopop_library.PatchApplicator.load import load_patches
from discopop_library.PatchApplicator.rollback import rollback_patches
from discopop_library.PathManagement.PathManagement import load_file_mapping


def run(arguments: PatchApplicatorArguments) -> int:
    """Return values:"
    "0: Applied successfully"
    "1: Nothing applied, error"
    "2: Some changes applied successfully"
    "3: Nothing to roll back, trivially successful"
    """

    if arguments.verbose:
        print("Started DiscoPoP Patch Applicator...")
        print("Working directory: ", os.getcwd())
        print(arguments)

    # create a directory for the patch applicator
    patch_applicator_dir = os.path.join(os.getcwd(), "patch_applicator")
    if not os.path.exists(patch_applicator_dir):
        if arguments.verbose:
            print("Creating patch_applicator directory...")
        os.mkdir(patch_applicator_dir)

    # create a file to store applied suggestions
    applied_suggestions_file = os.path.join(patch_applicator_dir, "applied_suggestions.json")
    if not os.path.exists(applied_suggestions_file):
        if arguments.verbose:
            print("Creating applied_suggestions.json file...")
        with open(applied_suggestions_file, "w+") as f:
            f.write(json.dumps({"applied": []}))

    # load file mapping
    file_mapping_path = os.path.join(os.getcwd(), "FileMapping.txt")
    if not os.path.exists(file_mapping_path):
        raise FileNotFoundError(
            "No file mapping found. Please execute the discopop_explorer in advance."
            + "\nExpected file: "
            + file_mapping_path
        )
    if arguments.verbose:
        print("Loading file mapping...")
    file_mapping = load_file_mapping(file_mapping_path)

    # check if patches have been generated
    patch_generator_dir = os.path.join(os.getcwd(), "patch_generator")
    if not os.path.exists(patch_generator_dir):
        raise FileNotFoundError(
            "No patches have been generated.\n"
            + "Please execute the discopop_patch_generator in advance.\n"
            + "Expected folder: "
            + patch_generator_dir
        )

    # handle arguments
    retval = 0
    if len(arguments.apply) > 0:
        retval = apply_patches(arguments.apply, file_mapping, arguments, applied_suggestions_file, patch_generator_dir)
    elif len(arguments.rollback) > 0:
        retval = rollback_patches(
            arguments.rollback, file_mapping, arguments, applied_suggestions_file, patch_generator_dir
        )
    elif arguments.clear:
        retval = clear_patches(file_mapping, arguments, applied_suggestions_file, patch_generator_dir)
    elif arguments.load:
        retval = load_patches(file_mapping, arguments, applied_suggestions_file, patch_generator_dir)
    elif arguments.list:
        print("Applied suggestions: ", list_applied_suggestions(arguments, applied_suggestions_file))

    if arguments.verbose:
        print("Done.")

    return retval



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, Dict, cast, Optional, Union

from discopop_explorer.PEGraphX import CUNode, NodeType, EdgeType, Node, PEGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    ParallelRegionInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import (
    line_contained_in_region,
    get_parent_of_type,
    get_cus_inside_function,
    check_reachability,
)
from discopop_explorer.utils import is_loop_index2


def filter_data_sharing_clauses(
    pet: PEGraphX, suggestions: List[PatternInfo], var_def_line_dict: Dict[str, List[str]]
) -> List[PatternInfo]:
    """Wrapper to filter data sharing clauses according to the included steps.
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    :return: List[PatternInfo]"""
    suggestions = __filter_data_sharing_clauses_by_function(pet, suggestions, var_def_line_dict)
    suggestions = __filter_data_sharing_clauses_by_scope(pet, suggestions, var_def_line_dict)
    suggestions = __filter_data_sharing_clauses_suppress_shared_loop_index(pet, suggestions)
    return suggestions


def __filter_data_sharing_clauses_suppress_shared_loop_index(pet: PEGraphX, suggestions: List[PatternInfo]):
    """Removes clauses for shared loop indices.
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :return: List[PatternInfo]"""
    for suggestion in suggestions:
        # only consider task suggestions
        if type(suggestion) != TaskParallelismInfo:
            continue
        if suggestion.type is not TPIType.TASK:
            continue
        # get parent loops of suggestion
        parent_loops_plus_last_node = get_parent_of_type(pet, suggestion._node, NodeType.LOOP, EdgeType.CHILD, True)
        parent_loops = [e[0] for e in parent_loops_plus_last_node]
        # consider only loops which enclose the suggestion
        parent_loops = [
            loop
            for loop in parent_loops
            if line_contained_in_region(suggestion._node.start_position(), loop.start_position(), loop.end_position())
        ]
        to_be_removed = []
        for var in suggestion.shared:
            for parent_loop in parent_loops:
                if is_loop_index2(pet, parent_loop, var):
                    to_be_removed.append(var)
        to_be_removed = list(set(to_be_removed))
        suggestion.shared = [v for v in suggestion.shared if v not in to_be_removed]
    return suggestions


def __filter_data_sharing_clauses_by_function(
    pet: PEGraphX, suggestions: List[PatternInfo], var_def_line_dict: Dict[str, List[str]]
) -> List[PatternInfo]:
    """Removes superfluous variables (not known in parent function of suggestion) from the data sharing clauses
    of task suggestions.
    Removes .addr suffix from variable names
    Removes entries if Variable occurs in different classes. Removes in following order:
    firstprivate, private, shared
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    :return: List[PatternInfo]
    """
    for suggestion in suggestions:
        # only consider task suggestions
        if type(suggestion) != TaskParallelismInfo:
            continue
        if suggestion.type not in [TPIType.TASK, TPIType.TASKLOOP]:
            continue
        # get function containing the task cu
        parent_function, last_node = get_parent_of_type(pet, suggestion._node, NodeType.FUNC, EdgeType.CHILD, True)[0]
        # filter firstprivate
        __filter_firstprivate_clauses(suggestion, parent_function, var_def_line_dict)
        # filter private
        __filter_private_clauses(suggestion, parent_function, var_def_line_dict)
        # filter shared
        __filter_shared_clauses(suggestion, parent_function, var_def_line_dict)

        # remove duplicates and .addr suffix from variable names
        suggestion.shared = list(set([v.replace(".addr", "") for v in suggestion.shared]))
        suggestion.private = list(set([v.replace(".addr", "") for v in suggestion.private]))
        suggestion.first_private = list(set([v.replace(".addr", "") for v in suggestion.first_private]))

        # remove duplicates (variable occurring in different classes)
        remove_from_first_private = []
        remove_from_private = []
        for var in suggestion.shared:
            if var in suggestion.private:
                remove_from_private.append(var)
            if var in suggestion.first_private:
                remove_from_first_private.append(var)
        for var in suggestion.private:
            if var in suggestion.first_private:
                remove_from_first_private.append(var)
        remove_from_first_private = list(set(remove_from_first_private))
        remove_from_private = list(set(remove_from_private))
        remove_from_private = [var for var in remove_from_private if var not in remove_from_first_private]
        suggestion.private = [var for var in suggestion.private if var not in remove_from_private]
        suggestion.first_private = [var for var in suggestion.first_private if var not in remove_from_first_private]
    return suggestions


def __filter_shared_clauses(suggestion: TaskParallelismInfo, parent_function, var_def_line_dict: Dict[str, List[str]]):
    """helper function for filter_data_sharing_clauses_by_function.
    Filters shared clauses.
    :param suggestion: Suggestion to be checked
    :param parent_function: parent function CU Node
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    """
    to_be_removed = []
    for var in suggestion.shared:
        var = var.replace(".addr", "")
        is_valid = False
        try:
            for def_line in var_def_line_dict[var]:
                # ensure backwards compatibility (no definition line present in cu_xml
                if def_line == "GlobalVar" or def_line == "LineNotFound":
                    is_valid = False
                    break
                if def_line is None:
                    is_valid = True
                # check if var is defined in parent function
                if line_contained_in_region(def_line, parent_function.start_position(), parent_function.end_position()):
                    is_valid = True
                else:
                    pass
        except ValueError as ve:
            raise ve
            pass
        if not is_valid:
            to_be_removed.append(var)
    to_be_removed = list(set(to_be_removed))
    suggestion.shared = [v for v in suggestion.shared if not v.replace(".addr", "") in to_be_removed]


def __filter_private_clauses(suggestion: TaskParallelismInfo, parent_function, var_def_line_dict: Dict[str, List[str]]):
    """helper function for filter_data_sharing_clauses_by_function.
    Filters private clauses.
    :param suggestion: Suggestion to be checked
    :param parent_function: parent function CU Node
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    """
    to_be_removed = []
    for var in suggestion.private:
        var = var.replace(".addr", "")
        is_valid = False
        try:
            for defLine in var_def_line_dict[var]:
                # catch GlobalVar and LineNotFound
                if defLine == "GlobalVar":
                    is_valid = True
                    continue
                if defLine == "LineNotFound":
                    continue
                # ensure backwards compatibility (no definition line present in cu_xml
                if defLine is None:
                    is_valid = True
                # check if var is defined in parent function
                elif line_contained_in_region(
                    defLine, parent_function.start_position(), parent_function.end_position()
                ):
                    is_valid = True
                else:
                    pass
        except ValueError:
            pass
        if not is_valid:
            to_be_removed.append(var)
    to_be_removed = list(set(to_be_removed))
    suggestion.private = [v for v in suggestion.private if not v.replace(".addr", "") in to_be_removed]


def __filter_firstprivate_clauses(
    suggestion: TaskParallelismInfo, parent_function, var_def_line_dict: Dict[str, List[str]]
):
    """helper function for filter_data_sharing_clauses_by_function.
    Filters firstprivate clauses.
    :param suggestion: Suggestion to be checked
    :param parent_function: parent function CU Node
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    """
    to_be_removed = []
    for var in suggestion.first_private:
        var = var.replace(".addr", "")
        is_valid = False
        try:
            for defLine in var_def_line_dict[var]:
                # ensure backwards compatibility (no definition line present in cu_xml
                if defLine is None:
                    is_valid = True
                # check if var is defined in parent function
                if line_contained_in_region(defLine, parent_function.start_position(), parent_function.end_position()):
                    is_valid = True
                else:
                    pass
        except ValueError:
            pass
        if not is_valid:
            to_be_removed.append(var)
    to_be_removed = list(set(to_be_removed))
    suggestion.first_private = [v for v in suggestion.first_private if not v.replace(".addr", "") in to_be_removed]


def __reverse_reachable_w_o_breaker(pet: PEGraphX, root: Node, target: Node, breaker_cu: Node, visited: List[Node]):
    """Helper function for filter_data_sharing_clauses_by_scope.
    Checks if target is reachable by traversing the successor graph in reverse, starting from root,
    without visiting breaker_cu.
    :param pet: PET Graph
    :param root: root node
    :param target: target node
    :param breaker_cu: breaker cu
    :param visited: list of already visited CUNodes
    :return: True, if target is reachable without visiting breaker_cu. False, otherwise."""
    if root in visited:
        return False
    visited.append(root)
    if root == target:
        return True
    if root == breaker_cu:
        return False
    recursion_result = False
    # start recursion for each incoming edge
    for tmp_e in pet.in_edges(root.id, EdgeType.SUCCESSOR):
        recursion_result = recursion_result or __reverse_reachable_w_o_breaker(
            pet, pet.node_at(tmp_e[0]), target, breaker_cu, visited
        )
    return recursion_result


def __filter_data_sharing_clauses_by_scope(
    pet: PEGraphX, suggestions: List[PatternInfo], var_def_line_dict: Dict[str, List[str]]
) -> List[PatternInfo]:
    """Filters out such data sharing clauses which belong to unknown variables at the source location of a given
    suggestion.
    Idea (per shared variable / suggestion):
    1.  Validate based on control flow:
        check if reverse path through successor graph from Task Suggestion CU to parent function CU exists,
        which doesn't cross the CU containing variable definition line (i.e. definition and task suggestion in separate
        branches of control flow graph).
    2.  Validate based on Scoping:
        If var-def has an incoming child-edge from X,
        check if task suggestions CU is reachable from X by traversing the child graph (i.e. TS CU is inside or
        below scope of X, thus var might be known).
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    :return: List[PatternInfo]
    """
    for suggestion in suggestions:
        # only consider task suggestions
        if type(suggestion) != TaskParallelismInfo:
            continue
        if suggestion.type is not TPIType.TASK:
            continue
        # get function containing the task cu
        parent_function_cu, last_node = get_parent_of_type(pet, suggestion._node, NodeType.FUNC, EdgeType.CHILD, True)[
            0
        ]
        # filter firstprivate
        __filter_sharing_clause(pet, suggestion, var_def_line_dict, parent_function_cu, "FP")
        # filter private
        __filter_sharing_clause(pet, suggestion, var_def_line_dict, parent_function_cu, "PR")
        # filter shared
        __filter_sharing_clause(pet, suggestion, var_def_line_dict, parent_function_cu, "SH")
    return suggestions


def __filter_sharing_clause(
    pet: PEGraphX,
    suggestion: TaskParallelismInfo,
    var_def_line_dict: Dict[str, List[str]],
    parent_function_cu,
    target_clause_list: str,
):
    """Helper function for filter_data_sharing_clauses_by_scope.
    Filters a given suggestions private, firstprivate or shared variables list,
    depending on the specific value of target_clause_list.
    :param pet: PET Graph
    :param suggestion: suggestion to be checked
    :param var_def_line_dict: dictionary containing: var_name -> [definition lines]
    :param parent_function_cu: CUNode corresponding to the function which contains the suggestion
    :param target_clause_list: One of: ['FP', 'PR', 'SH'], used to control which list of variables is filtered
    """
    to_be_removed = []
    if target_clause_list == "FP":
        sharing_clause_list = suggestion.first_private
    elif target_clause_list == "PR":
        sharing_clause_list = suggestion.private
    else:
        sharing_clause_list = suggestion.shared

    for var in sharing_clause_list:
        for var_def_line in var_def_line_dict[var]:
            if var_def_line == "GlobalVar":
                # accept global vars
                continue
            # get CU which contains var_def_line
            var_def_cu: Optional[Node] = None
            for child_cu in get_cus_inside_function(pet, parent_function_cu):
                if line_contained_in_region(var_def_line, child_cu.start_position(), child_cu.end_position()):
                    var_def_cu = child_cu
            if var_def_cu is None:
                continue
            # 1. check control flow (reverse BFS from suggestion._node to parent_function
            if __reverse_reachable_w_o_breaker(
                pet, pet.node_at(suggestion.node_id), parent_function_cu, var_def_cu, []
            ):
                # remove var as it may not be known
                to_be_removed.append(var)
                continue
            # 2. check if task suggestion is child of same nodes as var_def_cu
            for in_child_edge in pet.in_edges(var_def_cu.id, [EdgeType.CHILD, EdgeType.CALLSNODE]):
                parent_cu = pet.node_at(in_child_edge[0])
                # check if task suggestion cu is reachable from parent via child edges
                if not check_reachability(pet, suggestion._node, parent_cu, [EdgeType.CHILD]):
                    to_be_removed.append(var)
        to_be_removed = list(set(to_be_removed))
        if target_clause_list == "FP":
            suggestion.first_private = [v for v in suggestion.first_private if v not in to_be_removed]
        elif target_clause_list == "PR":
            suggestion.private = [v for v in suggestion.private if v not in to_be_removed]
        else:
            suggestion.shared = [v for v in suggestion.shared if v not in to_be_removed]


def remove_useless_barrier_suggestions(
    pet: PEGraphX, suggestions: List[TaskParallelismInfo]
) -> List[TaskParallelismInfo]:
    """remove suggested barriers which are not contained in the same
    function body with at least one suggested task.
    Returns the filtered version of the list given as a parameter.
    :param pet: PET graph
    :param suggestions: List[TaskParallelismInfo]
    :return: List[TaskParallelismInfo]
    """
    # split suggestions into task and taskwait suggestions
    taskwait_suggestions = []
    task_suggestions = []
    result_suggestions = []
    for single_suggestion in suggestions:
        if single_suggestion.type is TPIType.TASKWAIT:
            taskwait_suggestions.append(single_suggestion)
        elif single_suggestion.type is TPIType.TASK:
            task_suggestions.append(single_suggestion)
        else:
            result_suggestions.append(single_suggestion)
    # get map of function body cus containing task suggestions to line number
    # of task pragmas
    relevant_function_bodies: Dict[Node, List[str]] = {}
    for ts in task_suggestions:
        # get first parent cu with type function using bfs
        parent: Node = get_parent_of_type(pet, ts._node, NodeType.FUNC, EdgeType.CHILD, True)[0][0]
        if parent not in relevant_function_bodies:
            relevant_function_bodies[parent] = [ts.pragma_line]
        else:
            relevant_function_bodies[parent].append(ts.pragma_line)
    # remove suggested barriers which are no descendants of relevant functions
    result_suggestions += task_suggestions
    for tws in taskwait_suggestions:
        tws_line_number = tws.pragma_line
        tws_line_number = tws_line_number[tws_line_number.index(":") + 1 :]
        for rel_func_body in relevant_function_bodies.keys():
            if check_reachability(pet, tws._node, rel_func_body, [EdgeType.CHILD]):
                # remove suggested barriers where line number smaller than
                # pragma line number of task
                for line_number in relevant_function_bodies[rel_func_body]:
                    if line_number <= tws_line_number:
                        result_suggestions.append(tws)
                        break
    return result_suggestions


def remove_duplicate_data_sharing_clauses(suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """removes duplicates from in, out and in-out dependency lists.
    Mainly needed for printing purposes.
    :param suggestions: List[PatternInfo]
    :return: Modified List of PatternInfos
    """
    result: List[PatternInfo] = []
    for s in suggestions:
        if not type(s) == TaskParallelismInfo:
            result.append(s)
        else:
            s.in_dep = list(set(s.in_dep))
            s.out_dep = list(set(s.out_dep))
            s.in_out_dep = list(set(s.in_out_dep))
            result.append(s)
    return result


def __filter_in_dependencies(
    pet: PEGraphX,
    suggestion: TaskParallelismInfo,
    var_def_line_dict: Dict[str, List[str]],
    parent_function: Node,
    out_dep_vars: Dict[str, List[str]],
) -> bool:
    """Helper function for filter_data_depend_clauses.
    Filters in-dependencies of the given suggestion.
    :param pet: PET Graph
    :param suggestion: suggestion to be filtered
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    :param parent_function: parent function containing suggestion
    :param out_dep_vars: variables used in outgoing dependencies
    :return: True, if a modification of suggestion.out_dep has been made. False, otherwise."""
    to_be_removed = []
    modification_found = False
    for var in suggestion.in_dep:
        var = var.replace(".addr", "")
        is_valid = False
        try:
            for defLine in var_def_line_dict[var]:
                # ensure backwards compatibility (no definition line present in cu_xml
                if defLine is None:
                    is_valid = True
                # check if var is defined in parent function
                if line_contained_in_region(defLine, parent_function.start_position(), parent_function.end_position()):
                    # check if var is contained in out_dep_vars and a previous out_dep exists
                    if var in out_dep_vars:
                        for line_num in out_dep_vars[var]:
                            line_num = str(line_num)
                            if ":" in line_num:
                                line_num = line_num.split(":")[1]
                            # check validity of the dependence by reachability checking on
                            # successor + child graph

                            # get CU containing line_num
                            for cu_node in pet.all_nodes(CUNode):
                                file_id = suggestion._node.start_position().split(":")[0]
                                test_line = file_id + ":" + line_num
                                # check if line_num is contained in cu_node
                                if not line_contained_in_region(
                                    test_line, cu_node.start_position(), cu_node.end_position()
                                ):
                                    continue
                                # check if path from cu_node to suggestion._node exists
                                if check_reachability(
                                    pet,
                                    suggestion._node,
                                    cu_node,
                                    [EdgeType.SUCCESSOR, EdgeType.CHILD],
                                ):
                                    is_valid = True
                else:
                    pass
        except ValueError:
            pass
        if not is_valid:
            modification_found = True
            to_be_removed.append(var)
    to_be_removed = list(set(to_be_removed))
    suggestion.in_dep = [v for v in suggestion.in_dep if not v.replace(".addr", "") in to_be_removed]
    return modification_found


def __filter_out_dependencies(
    pet: PEGraphX,
    suggestion: TaskParallelismInfo,
    var_def_line_dict: Dict[str, List[str]],
    parent_function: Node,
    in_dep_vars: Dict[str, List[str]],
) -> bool:
    """Helper function for filter_data_depend_clauses.
    Filters out-dependencies of the given suggestion.
    :param pet: PET Graph
    :param suggestion: suggestion to be filtered
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    :param parent_function: parent function containing suggestion
    :param in_dep_vars: variables used in incoming dependencies
    :return: True, if a modification of suggestion.out_dep has been made. False, otherwise."""
    modification_found = False
    to_be_removed = []
    for var in suggestion.out_dep:
        var = var.replace(".addr", "")
        is_valid = False
        try:
            for defLine in var_def_line_dict[var]:
                # ensure backwards compatibility (no definition line present in cu_xml
                if defLine is None:
                    is_valid = True
                # check if var is defined in parent function
                if line_contained_in_region(defLine, parent_function.start_position(), parent_function.end_position()):
                    # check if var is contained in in_dep_vars and a successive in_dep exists
                    if var in in_dep_vars:
                        for line_num in in_dep_vars[var]:
                            line_num = str(line_num)
                            if ":" in line_num:
                                line_num = line_num.split(":")[1]
                            # check validity of the dependence by reachability checking on
                            # successor + child graph

                            # get CU containing line_num
                            for cu_node in pet.all_nodes(CUNode):
                                file_id = suggestion._node.start_position().split(":")[0]
                                test_line = file_id + ":" + line_num
                                # check if line_num is contained in cu_node
                                if not line_contained_in_region(
                                    test_line, cu_node.start_position(), cu_node.end_position()
                                ):
                                    continue
                                # check if path from suggestion._node to cu_node exists
                                if check_reachability(
                                    pet,
                                    cu_node,
                                    suggestion._node,
                                    [EdgeType.SUCCESSOR, EdgeType.CHILD],
                                ):
                                    is_valid = True
                else:
                    pass
        except ValueError:
            pass
        if not is_valid:
            to_be_removed.append(var)
            modification_found = True
    to_be_removed = list(set(to_be_removed))
    suggestion.out_dep = [v for v in suggestion.out_dep if not v.replace(".addr", "") in to_be_removed]
    return modification_found


def __filter_in_out_dependencies(
    pet: PEGraphX,
    suggestion: TaskParallelismInfo,
    var_def_line_dict: Dict[str, List[str]],
    parent_function: Node,
    in_dep_vars: Dict[str, List[str]],
    out_dep_vars: Dict[str, List[str]],
) -> bool:
    """Helper function for filter_data_depend_clauses.
    Filters in_out-dependencies of the given suggestion.
    :param pet: PET Graph
    :param suggestion: suggestion to be filtered
    :param var_def_line_dict: Dictionary containing: var_name -> [definition lines]
    :param parent_function: parent function containing suggestion
    :param in_dep_vars: variables used in incoming dependencies
    :param out_dep_vars: variables used in incoming dependencies
    :return: True, if a modification of suggestion.in_out_dep has been made. False, otherwise."""
    modification_found = False
    to_be_removed = []
    for var in suggestion.in_out_dep:
        var = var.replace(".addr", "")
        is_valid = False
        try:
            for defLine in var_def_line_dict[var]:
                # ensure backwards compatibility (no definition line present in cu_xml
                if defLine is None:
                    is_valid = True
                # check if var is defined in parent function
                if line_contained_in_region(defLine, parent_function.start_position(), parent_function.end_position()):
                    # check if var occurs more than once as in or out, i.e. at least an actual in or out
                    # dependency exists
                    if len(in_dep_vars[var]) > 1 or len(out_dep_vars[var]) > 1:
                        # check if out dep prior an in dep afterwards exist
                        prior_out_exists = False
                        successive_in_exists = False
                        for line_num in out_dep_vars[var]:
                            line_num = str(line_num)
                            if ":" in line_num:
                                line_num = line_num.split(":")[1]
                            # check validity of the dependence by reachability checking on
                            # successor + child graph

                            # get CU containing line_num
                            for cu_node in pet.all_nodes(CUNode):
                                file_id = suggestion._node.start_position().split(":")[0]
                                test_line = file_id + ":" + line_num
                                # check if line_num is contained in cu_node
                                if not line_contained_in_region(
                                    test_line, cu_node.start_position(), cu_node.end_position()
                                ):
                                    continue
                                # check if path from cu_node to suggestion._node exists
                                if check_reachability(
                                    pet,
                                    suggestion._node,
                                    cu_node,
                                    [EdgeType.SUCCESSOR, EdgeType.CHILD],
                                ):
                                    prior_out_exists = True
                        for line_num in in_dep_vars[var]:
                            line_num = str(line_num)
                            if ":" in line_num:
                                line_num = line_num.split(":")[1]
                            # check validity of the dependence by reachability checking on
                            # successor + child graph

                            # get CU containing line_num
                            for cu_node in pet.all_nodes(CUNode):
                                file_id = suggestion._node.start_position().split(":")[0]
                                test_line = file_id + ":" + line_num
                                # check if line_num is contained in cu_node
                                if not line_contained_in_region(
                                    test_line, cu_node.start_position(), cu_node.end_position()
                                ):
                                    continue
                                # check if path from suggestion._node to cu_node exists
                                if check_reachability(
                                    pet,
                                    cu_node,
                                    suggestion._node,
                                    [EdgeType.SUCCESSOR, EdgeType.CHILD],
                                ):
                                    successive_in_exists = True
                        # check and treat conditions
                        if prior_out_exists and successive_in_exists:
                            # proper in_out_dep
                            is_valid = True
                        elif prior_out_exists and not successive_in_exists:
                            # depend in
                            suggestion.in_dep.append(var)
                            suggestion.in_dep = list(set(suggestion.in_dep))
                        elif not prior_out_exists and successive_in_exists:
                            # depend out
                            suggestion.out_dep.append(var)
                            suggestion.out_dep = list(set(suggestion.out_dep))
                else:
                    pass
        except ValueError:
            pass
        if not is_valid:
            to_be_removed.append(var)
            modification_found = True
    to_be_removed = list(set(to_be_removed))
    suggestion.in_out_dep = [v for v in suggestion.in_out_dep if not v.replace(".addr", "") in to_be_removed]
    return modification_found


def filter_data_depend_clauses(
    pet: PEGraphX, suggestions: List[PatternInfo], var_def_line_dict: Dict[str, List[str]]
) -> List[PatternInfo]:
    """Removes superfluous variables from the data depend clauses
    of task suggestions.
    :param pet: PET graph
    :param suggestions: List[PatternInfo]
    :param var_def_line_dict: Dictionary containing mapping from var_name to [definition lines]
    :return: List[PatternInfo]
    """
    modification_found = True
    while modification_found:
        # get list of used variabled by dependency type
        in_dep_vars: Dict[str, List[str]] = dict()
        out_dep_vars: Dict[str, List[str]] = dict()
        for suggestion in suggestions:
            # only consider task suggestions
            if type(suggestion) != TaskParallelismInfo:
                continue
            if suggestion.type not in [TPIType.TASK, TPIType.TASKLOOP]:
                continue
            for var in suggestion.in_dep:
                if var not in in_dep_vars:
                    in_dep_vars[var] = []
                in_dep_vars[var].append(suggestion.pragma_line)
            for var in suggestion.out_dep:
                if var not in out_dep_vars:
                    out_dep_vars[var] = []
                out_dep_vars[var].append(suggestion.pragma_line)
            for var in suggestion.in_out_dep:
                if var not in in_dep_vars:
                    in_dep_vars[var] = []
                in_dep_vars[var].append(suggestion.pragma_line)
                if var not in out_dep_vars:
                    out_dep_vars[var] = []
                out_dep_vars[var].append(suggestion.pragma_line)

        # filter dependency clauses
        modification_found = False
        for suggestion in suggestions:
            # only consider task suggestions
            if type(suggestion) != TaskParallelismInfo:
                continue
            if suggestion.type not in [TPIType.TASK, TPIType.TASKLOOP]:
                continue
            # get function containing the task cu
            parent_function, last_node = get_parent_of_type(pet, suggestion._node, NodeType.FUNC, EdgeType.CHILD, True)[
                0
            ]
            # filter in_dep
            modification_found = modification_found or __filter_in_dependencies(
                pet, suggestion, var_def_line_dict, parent_function, out_dep_vars
            )
            # filter out_dep
            modification_found = modification_found or __filter_out_dependencies(
                pet, suggestion, var_def_line_dict, parent_function, in_dep_vars
            )
            # filter in_out_dep
            modification_found = modification_found or __filter_in_out_dependencies(
                pet, suggestion, var_def_line_dict, parent_function, in_dep_vars, out_dep_vars
            )

            # correct in_out_vars (find in_out vars if not already detected)
            overlap = [v for v in suggestion.in_dep if v in suggestion.out_dep]
            for v in overlap:
                if v not in suggestion.in_out_dep:
                    modification_found = True
                suggestion.in_dep.remove(v)
                suggestion.out_dep.remove(v)
                suggestion.in_out_dep.append(v)
            suggestion.in_out_dep = list(set(suggestion.in_out_dep))
            # correct in_out vars (remove from in and out)
            for v in suggestion.in_out_dep:
                if v in suggestion.in_dep:
                    suggestion.in_dep.remove(v)
                    modification_found = True
                if v in suggestion.out_dep:
                    suggestion.out_dep.remove(v)
                    modification_found = True

    return suggestions


def remove_duplicates(suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """removes duplicates from the list of suggestions and return the modified
    list.
    CU-ID is not considered.
    Removes a suggestion, if one with identical region_start_line,
    region_end_line and pragma exists.
    :param suggestions: List[PatternInfo]
    :return: List[PatternInfo]
    """
    buffer = []  # list of tuples containing region_start_line,
    # region_end_line and pragma, representing suggestions
    result = []
    for sug in suggestions:
        if type(sug) == ParallelRegionInfo:
            sug_tmp = cast(Union[TaskParallelismInfo, ParallelRegionInfo], sug)
        elif type(sug) == TaskParallelismInfo:
            sug_tmp = cast(Union[TaskParallelismInfo, ParallelRegionInfo], sug)
        else:
            continue
        representing_tuple = (sug_tmp.region_start_line, sug_tmp.region_end_line, sug_tmp.pragma)
        if representing_tuple in buffer:
            continue
        else:
            buffer.append(representing_tuple)
            result.append(sug)
    return result



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter as tk


class ToolTip(object):
    # based on https://stackoverflow.com/questions/20399243/display-message-when-hovering-over-something-with-mouse-cursor-in-python

    def __init__(self, widget):
        self.widget = widget
        self.tip_window = None
        self.id = None
        self.x = self.y = 0
        self.text = ""

    def showtip(self, text):
        """Display text in tooltip window"""
        self.text = text
        if self.tip_window or not self.text:
            return
        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 57
        y = y + cy + self.widget.winfo_rooty() + 27
        self.tip_window = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(
            tw,
            text=self.text,
            justify=tk.LEFT,
            background="#ffffe0",
            relief=tk.SOLID,
            borderwidth=1,
            font=("tahoma", "8", "normal"),
        )
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tip_window
        self.tip_window = None
        if tw:
            tw.destroy()


def create_tool_tip(widget, text):
    tool_tip = ToolTip(widget)

    def enter(event):
        tool_tip.showtip(text)

    def leave(event):
        tool_tip.hidetip()

    widget.bind("<Enter>", enter)
    widget.bind("<Leave>", leave)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os
from typing import Dict


def save_line_mapping(line_mapping_dict: Dict[str, Dict[str, int]], discopop_path: str = ""):
    """dumps line_mapping_dict to line_mapping.json"""
    # dump line mapping to json file
    with open(os.path.join(discopop_path, "line_mapping.json"), "w+") as f:
        json.dump(line_mapping_dict, f)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import random
import string
import subprocess
import tkinter as tk
from json import JSONDecodeError
from tkinter import filedialog
from typing import TextIO, List, Dict, Tuple, Any

import jsons  # type:ignore

from DEPRECATED_discopop_wizard.screens.execution import ExecutionView
from DEPRECATED_discopop_wizard.screens.optimizer.binding import create_optimizer_screen
from DEPRECATED_discopop_wizard.screens.suggestions.overview import (
    show_suggestions_overview_screen,
    get_suggestion_objects,
)
from DEPRECATED_discopop_wizard.screens.utils import create_tool_tip


class ExecutionConfiguration(object):
    button: tk.Button
    value_dict: Dict[str, Any]

    def __init__(self, wizard):
        self.value_dict = {
            "label": "",
            "description": "",
            "executable_name": "",
            "executable_arguments": "",
            "make_flags": "",
            "project_path": "",
            "linker_flags": "",
            "make_target": "",
            "memory_profiling_skip_function_parameters": 0,
            "notes": "",
            "working_copy_path": "",
            "tags": "",
            "explorer_flags": "--json=patterns.json --dump-detection-result",
        }
        self.value_dict["id"] = "".join(random.choices(string.ascii_uppercase + string.digits, k=8))
        self.wizard = wizard

        # self.__extract_values_from_help_string()

    def __extract_values_from_help_string(self):
        """Read help strings of:
        - DiscoPoP Explorer
        - runDiscoPoP script
        in order to automatically determine the required and optional arguments"""
        targets: List[Tuple[str, str]] = [  # [(name, executable)]
            ("rundiscopop", self.wizard.settings.discopop_build_dir + "/scripts/runDiscoPoP"),
            ("explorer", "discopop_explorer"),
        ]
        for target_name, target_executable in targets:
            # get help string
            tmp = subprocess.run([target_executable, "--help"], stdout=subprocess.PIPE)
            help_string = tmp.stdout.decode("utf-8")
            for line in help_string.split("\n"):
                line = line.replace("\t", " ")
                while "  " in line:
                    line = line.replace("  ", " ")
                if line.startswith(" "):
                    line = line[1:]
                if not line.startswith("* --"):
                    continue
                line = line[4:]  # cut off '* --'
                # extract name and type
                name_cutoff_index = len(line)
                if " " in line:
                    name_cutoff_index = min(name_cutoff_index, line.index(" "))
                if "=" in line:
                    name_cutoff_index = min(name_cutoff_index, line.index("="))
                value_name = line[:name_cutoff_index]
                if value_name == "help":
                    continue
                line = line[name_cutoff_index:]
                # extract type if present
                value_type = "bool"
                if "<" in line and ">" in line:
                    left_index = line.index("<") + 1
                    right_index = line.index(">")
                    value_type = line[left_index:right_index]
                # add value to dictionaries
                self.value_dict[target_name + "_" + value_name] = ""
                # self.type_dict[target_name + "_" + value_name] = value_type

    def get_as_button(self, wizard, main_screen_obj, parent_frame: tk.Frame, all_buttons: List[tk.Button]) -> tk.Button:
        button = tk.Button(parent_frame, text=self.value_dict["label"])
        button.config(
            command=lambda: self.highlight_and_update_notebook_screens(wizard, main_screen_obj, button, all_buttons)
        )
        self.button = button
        return button

    def init_from_json(self, json_file: TextIO):
        json_str = ""
        for line in json_file.readlines():
            json_str += line
        self.value_dict = {**self.value_dict, **jsons.loads(json_str)}  # merge both dictionaries

    def get_values_as_json_string(self):
        """returns a representation of the settings which will be stored in a configuration file."""
        return jsons.dumps(self.value_dict)

    def highlight_and_update_notebook_screens(
        self, wizard, main_screen_obj, pressed_button: tk.Button, all_buttons: List[tk.Button]
    ):
        # remove previous highlights
        for configuration_button in all_buttons:
            configuration_button.configure(state=tk.NORMAL)

        # highlight pressed configuration button
        pressed_button.configure(state=tk.DISABLED)

        # update details screen of pressed configuration button
        self.show_details_screen(wizard, main_screen_obj)

        # update results screen of pressed configuration button and set results tab state based on result existence
        main_screen_obj.notebook.tab(main_screen_obj.results_frame, state=self.__button_state_from_result_existence())
        main_screen_obj.notebook.tab(main_screen_obj.optimizer_frame, state=self.__button_state_from_result_existence())

        if self.__button_state_from_result_existence() == "normal":
            show_suggestions_overview_screen(wizard, main_screen_obj.results_frame, self)
        if self.__button_state_from_result_existence() == "normal":
            create_optimizer_screen(wizard, main_screen_obj.optimizer_frame, self)

    def show_details_screen(self, wizard, main_screen_obj):
        # delete previous frame contents
        for c in main_screen_obj.details_frame.winfo_children():
            c.destroy()

        frame = tk.Frame(main_screen_obj.details_frame)
        frame.grid(row=1, column=2)

        canvas = tk.Canvas(frame)
        canvas.grid(row=1)

        # show labels
        tk.Label(canvas, text="Label:", justify=tk.RIGHT, anchor="e", font=wizard.style_font_bold_small).grid(
            row=1, column=1, sticky="ew"
        )
        tk.Label(canvas, text="Description", justify=tk.RIGHT, anchor="e").grid(row=2, column=1, sticky="ew")
        tk.Label(
            canvas,
            text="Executable name:",
            justify=tk.RIGHT,
            anchor="e",
            font=wizard.style_font_bold_small,
        ).grid(row=3, column=1, sticky="ew")
        tk.Label(canvas, text="Executable arguments:", justify=tk.RIGHT, anchor="e").grid(row=4, column=1, sticky="ew")
        tk.Label(canvas, text="Make flags:", justify=tk.RIGHT, anchor="e").grid(row=5, column=1, sticky="ew")
        tk.Label(
            canvas,
            text="Project path:",
            justify=tk.RIGHT,
            anchor="e",
            font=wizard.style_font_bold_small,
        ).grid(row=6, column=1, sticky="ew")
        tk.Label(canvas, text="Project linker flags:", justify=tk.RIGHT, anchor="e").grid(row=7, column=1, sticky="ew")
        tk.Label(canvas, text="Make target:", justify=tk.RIGHT, anchor="e").grid(row=8, column=1, sticky="ew")
        tk.Label(
            canvas,
            text="Memory Profiling:",
            justify=tk.RIGHT,
            anchor="e",
            font=wizard.style_font_bold_small,
        ).grid(row=9, column=1, sticky="ew")
        tk.Label(canvas, text="Skip function params:", justify=tk.RIGHT, anchor="e").grid(row=10, column=1, sticky="ew")
        tk.Label(
            canvas,
            text="Additional:",
            justify=tk.RIGHT,
            anchor="e",
            font=wizard.style_font_bold_small,
        ).grid(row=11, column=1, sticky="ew")
        tk.Label(canvas, text="Tags:", justify=tk.RIGHT, anchor="e").grid(row=12, column=1, sticky="ew")
        tk.Label(canvas, text="Notes:", justify=tk.RIGHT, anchor="e").grid(row=13, column=1, sticky="ew")

        # show input fields
        label = tk.Entry(canvas)
        label.grid(row=1, column=2, sticky="ew")
        label.insert(tk.END, self.value_dict["label"])
        create_tool_tip(label, "Name of the configuration. Used to distinguish configurations in the main menu.")

        description = tk.Entry(canvas)
        description.grid(row=2, column=2, sticky="ew")
        description.insert(tk.END, self.value_dict["description"])

        executable_name = tk.Entry(canvas)
        executable_name.insert(tk.END, self.value_dict["executable_name"])
        executable_name.grid(row=3, column=2, sticky="ew")
        create_tool_tip(
            executable_name,
            "Name of the executable which is created when building the target project. The name will be "
            "used to execute the configuration.",
        )

        executable_args = tk.Entry(canvas)
        executable_args.grid(row=4, column=2, sticky="ew")
        executable_args.insert(tk.END, self.value_dict["executable_arguments"])
        create_tool_tip(
            executable_args,
            "Specify arguments which shall be forwarded to the call of the created executable for the " "profiling.",
        )

        make_flags = tk.Entry(canvas)
        make_flags.grid(row=5, column=2, sticky="ew")
        make_flags.insert(tk.END, str(self.value_dict["make_flags"]))
        create_tool_tip(
            make_flags,
            "Specified flags will be forwarded to Make during the build of the target project.",
        )

        project_path = tk.Entry(canvas)
        project_path.grid(row=6, column=2, sticky="ew")
        project_path.insert(tk.END, self.value_dict["project_path"])
        create_tool_tip(project_path, "Path to the project which shall be analyzed for potential parallelism.")

        def overwrite_with_selection(target: tk.Entry):
            prompt_result = tk.filedialog.askdirectory()
            if len(prompt_result) != 0:
                target.delete(0, tk.END)
                target.insert(0, prompt_result)

        project_path_selector = tk.Button(canvas, text="Select", command=lambda: overwrite_with_selection(project_path))
        project_path_selector.grid(row=6, column=3)

        project_linker_flags = tk.Entry(canvas)
        project_linker_flags.grid(row=7, column=2, sticky="ew")
        project_linker_flags.insert(tk.END, self.value_dict["linker_flags"])
        create_tool_tip(
            project_linker_flags,
            "Linker flags which need to be passed to the build system in order to create a valid " "executable.",
        )

        make_target = tk.Entry(canvas)
        make_target.grid(row=8, column=2, sticky="ew")
        make_target.insert(tk.END, self.value_dict["make_target"])
        create_tool_tip(make_target, "Space-separated list of make targets to be created.")

        mpsfp_var = tk.IntVar()
        mpsfp_var.set(self.value_dict["memory_profiling_skip_function_parameters"])
        memory_profiling_skip_function_parameters = tk.Checkbutton(canvas, onvalue=1, offvalue=0, variable=mpsfp_var)
        memory_profiling_skip_function_parameters.grid(row=10, column=2, sticky="w")
        create_tool_tip(
            memory_profiling_skip_function_parameters,
            "Disables the memory profiling for function arguments.\n\n"
            "Depending on the application, this may lead to significant profiling runtime\n"
            "improvements, but the correctness of the results can not be guaranteed anymore!\n"
            "Use this mode with caution, and be aware of potential issues!",
        )

        tags = tk.Entry(canvas)
        tags.grid(row=12, column=2, sticky="ew")
        tags.insert(tk.END, self.value_dict["tags"])
        create_tool_tip(tags, "Space-separated list of tags for identification.")

        additional_notes = tk.Text(canvas, height=10)
        additional_notes.grid(row=13, column=2, sticky="ew")
        additional_notes.insert(tk.END, self.value_dict["notes"])
        create_tool_tip(additional_notes, "Can be used to store notes regarding the configuration.")

        # show buttons
        button_canvas = tk.Canvas(frame)
        button_canvas.grid(row=2)
        # Create "Save" button
        save_button = tk.Button(
            button_canvas,
            text="Save",
            command=lambda: self.save_changes(
                wizard,
                main_screen_obj,
                label,
                description,
                executable_name,
                executable_args,
                make_flags,
                project_path,
                project_linker_flags,
                make_target,
                mpsfp_var,
                tags,
                additional_notes,
            ),
        )
        save_button.grid(row=1, column=1)

        # Create "Delete" button
        delete_button = tk.Button(
            button_canvas,
            text="Delete",
            command=lambda: self.delete_configuration(wizard, main_screen_obj, frame),
        )
        delete_button.grid(row=1, column=2)

        # Create "Open Folder" button
        if os.path.exists(os.path.join(project_path.get(), ".discopop")):
            target_path = os.path.join(project_path.get(), ".discopop")
        else:
            target_path = project_path.get()
        open_project_folder_button = tk.Button(
            button_canvas,
            text="Open Folder",
            command=lambda: os.system(
                "xdg-open "
                + str(
                    os.path.join(project_path.get(), ".discopop")
                    if os.path.exists(os.path.join(project_path.get(), ".discopop"))
                    else project_path.get()
                )
            ),
        )
        open_project_folder_button.grid(row=1, column=3)

        # Create "Execute" button
        execute_button = tk.Button(
            button_canvas,
            text="Execute",
            command=lambda: self.execute_configuration(
                wizard,
                main_screen_obj,
                label,
                description,
                executable_name,
                executable_args,
                make_flags,
                project_path,
                project_linker_flags,
                make_target,
                mpsfp_var,
                tags,
                additional_notes,
            ),
        )
        execute_button.grid(row=1, column=4)

    def __button_state_from_result_existence(self) -> str:
        # check if suggestions can be loaded. If so, enable the button.
        # Else, disable it.
        try:
            suggestions = get_suggestion_objects(self.wizard, self)
        except FileNotFoundError:
            return "disabled"
        except JSONDecodeError:
            return "disabled"
        return "normal"

    def save_changes(
        self,
        wizard,
        main_screen_obj,
        label,
        description,
        executable_name,
        executable_args,
        make_flags,
        project_path,
        project_linker_flags,
        make_target,
        memory_profiling_skip_function_parameters,
        tags,
        additional_notes,
        rebuild_configurations_frame=True,
    ):
        # update execution configuration
        self.value_dict["label"] = label.get()
        self.value_dict["description"] = description.get()
        self.value_dict["executable_name"] = executable_name.get()
        self.value_dict["executable_arguments"] = executable_args.get()
        self.value_dict["make_flags"] = make_flags.get()
        self.value_dict["project_path"] = project_path.get()
        self.value_dict["working_copy_path"] = self.value_dict["project_path"] + "/.discopop"
        self.value_dict["linker_flags"] = project_linker_flags.get()
        self.value_dict["make_target"] = make_target.get()
        self.value_dict["memory_profiling_skip_function_parameters"] = memory_profiling_skip_function_parameters.get()
        self.value_dict["tags"] = tags.get()
        self.value_dict["notes"] = additional_notes.get("1.0", tk.END)

        # construct config path
        config_path = os.path.join(
            wizard.config_dir,
            "execution_configurations",
            str(self.value_dict["id"]) + "_" + self.value_dict["label"] + ".json",
        )
        # remove old config if present
        if os.path.exists(config_path):
            os.remove(config_path)
        # write config to file
        with open(config_path, "w+") as f:
            f.write(self.get_values_as_json_string())

        if rebuild_configurations_frame:  # used to prevent de-selecting button on execution
            main_screen_obj.build_configurations_frame(wizard)
        print("Saved configuration")
        self.wizard.console.print("Saved configuration")

    def delete_configuration(self, wizard, main_screen_obj, details_frame: tk.Frame):
        # delete configuration file if it exists
        config_path = os.path.join(
            wizard.config_dir,
            "execution_configurations",
            str(self.value_dict["id"]) + "_" + self.value_dict["label"] + ".json",
        )
        if os.path.exists(config_path):
            os.remove(config_path)

        main_screen_obj.build_configurations_frame(wizard)
        # remove details view
        for c in details_frame.winfo_children():
            c.destroy()

    def execute_configuration(
        self,
        wizard,
        main_screen_obj,
        label,
        description,
        executable_name,
        executable_args,
        make_flags,
        project_path,
        project_linker_flags,
        make_target,
        memory_profiling_skip_function_parameters,
        tags,
        additional_notes,
    ):
        # save changes
        self.save_changes(
            wizard,
            main_screen_obj,
            label,
            description,
            executable_name,
            executable_args,
            make_flags,
            project_path,
            project_linker_flags,
            make_target,
            memory_profiling_skip_function_parameters,
            tags,
            additional_notes,
            rebuild_configurations_frame=False,
        )

        # create execution view and update results frame
        ExecutionView(self, wizard, main_screen_obj.results_frame)
        # set results tab state based on result existence
        main_screen_obj.notebook.tab(main_screen_obj.results_frame, state=self.__button_state_from_result_existence())
        main_screen_obj.notebook.tab(main_screen_obj.optimizer_frame, state=self.__button_state_from_result_existence())

        # show results tab
        main_screen_obj.notebook.select(main_screen_obj.results_frame)
        # initialize optimizer tab
        create_optimizer_screen(wizard, main_screen_obj.optimizer_frame, self)

    def get_tags(self) -> List[str]:
        """Returns a list of strings which represents the tags assigned to the configuration."""
        return [tag for tag in self.value_dict["tags"].split(" ") if len(tag) != 0]



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Tuple, List, Optional, cast

from sympy import Expr, Symbol
from sympy import Integer

from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum


class Device(object):
    __frequency: Expr
    __thread_count: Expr
    openmp_device_id: int

    def __init__(
        self,
        frequency: Expr,  # GHz
        thread_count: Expr,
        openmp_device_id: int,
        device_specific_compiler_flags: str,
    ):
        self.__frequency = frequency
        self.__thread_count = thread_count
        self.openmp_device_id = openmp_device_id
        self.device_specific_compiler_flags: str = device_specific_compiler_flags

    def get_device_specific_pattern_info(
        self, suggestion: PatternInfo, suggestion_type: str
    ) -> Tuple[PatternInfo, str]:
        return suggestion, suggestion_type

    def get_compute_capability(self) -> Expr:
        return self.__frequency

    def get_thread_count(self) -> Expr:
        return self.__thread_count

    def get_free_symbols(self) -> List[Tuple[Symbol, Optional[Expr]]]:
        result_list: List[Tuple[Symbol, Optional[Expr]]] = []
        result_list += [(cast(Symbol, s), None) for s in self.__frequency.free_symbols]
        result_list += [(cast(Symbol, s), None) for s in self.__thread_count.free_symbols]
        return result_list

    def get_estimated_execution_time_in_micro_seconds(self, workload: Expr, is_sequential: bool):
        """execution time is estimated by:
        - convert workload to estimated amount of CPU instructions using a extra-p model
        - NOTE: use "perf stat ./<cmd" to get the amount of instructions and instructions per cycle




        execution_time_in_micro_seconds = execution_time_in_seconds * 1000000
         execution_time_in_seconds = workload / instructions_per_second
         instructions_per_second = instructions_per_core_per_second * self.__thread_count
         instructions_per_core_per_second = self.__frequency / avg_cycles_per_instruction
         avg_cycles_per_instruction = 1 / avg_instructions_per_cycle
        """

        # todo -> define functions to evaluate workloads on different devices. include these functions in the construction of the cost model for later evaluation instead of "manually" converting the workload to time values
        # todo: correctly set is_sequential argument
        # todo mark parallel execution in subtree of suggestions?

        avg_instructions_per_cycle = 0.89  # todo (get from benchmarking)
        # current value determined by manual "measurement" based on a single example!
        average_cycles_per_instruction = 1 / avg_instructions_per_cycle
        instructions_per_core_per_second = self.__frequency / average_cycles_per_instruction
        instructions_per_second = instructions_per_core_per_second * (
            Integer(1) if is_sequential else self.__thread_count
        )
        workload_in_instructions = workload * 2.120152292  # todo (get from benchmarking / extra-p model)
        # current factor determined by manual "measurement" based on a single example!

        execution_time_in_seconds = workload_in_instructions / instructions_per_second
        execution_time_in_micro_seconds = execution_time_in_seconds * 1000000
        return execution_time_in_micro_seconds

    def get_device_type(self) -> DeviceTypeEnum:
        raise ValueError("This method need to be overwritten by subclasses!")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


import functools
from enum import Enum
from sys import platform


class Platform(Enum):
    UNKNOWN = 0
    LINUX = 1
    WINDOWS = 2
    OSX = 3


def get_platform():
    platforms = {
        "linux1": Platform.LINUX,
        "linux2": Platform.LINUX,
        "darwin": Platform.OSX,
        "win32": Platform.WINDOWS,
    }
    if platform not in platforms:
        return Platform.UNKNOWN

    return platforms[platform]


def support_scrolling(canvas):
    # add support for mouse wheel scrolling (on linux systems)
    def _on_mousewheel(event, scroll):
        canvas.yview_scroll(int(scroll), "units")

    def _bind_to_mousewheel(event):
        pf = get_platform()
        if pf == Platform.LINUX:
            canvas.bind_all("<Button-4>", functools.partial(_on_mousewheel, scroll=-1))
            canvas.bind_all("<Button-5>", functools.partial(_on_mousewheel, scroll=1))
        elif pf == Platform.WINDOWS:
            canvas.bind_all("<MouseWheel>", functools.partial(_on_mousewheel, scroll=(-1 * (event.delta / 120))))
        elif pf == Platform.OSX:
            canvas.bind_all("<MouseWheel>", functools.partial(_on_mousewheel, scroll=-1 * event.delta))
        else:
            canvas.bind_all("<Button-4>", functools.partial(_on_mousewheel, scroll=-1))
            canvas.bind_all("<Button-5>", functools.partial(_on_mousewheel, scroll=1))

    def _unbind_from_mousewheel(event):
        pf = get_platform()
        if pf == Platform.LINUX:
            canvas.unbind_all("<Button-4>")
            canvas.unbind_all("<Button-5>")
        elif pf == Platform.WINDOWS or pf == Platform.OSX:
            canvas.unbind_all("<MouseWheel>")
        else:
            canvas.unbind_all("<Button-4>")
            canvas.unbind_all("<Button-5>")

    canvas.bind("<Enter>", _bind_to_mousewheel)
    canvas.bind("<Leave>", _unbind_from_mousewheel)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List

from discopop_explorer.PEGraphX import EdgeType, FunctionNode, PEGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import line_contained_in_region


def suggest_shared_clauses_for_all_tasks_in_function_body(
    pet: PEGraphX, suggestions: List[PatternInfo]
) -> List[PatternInfo]:
    """Marks unmentioned variables as shared, if they occur as shared in a different task suggestions
    inside the parent functions body.
    :param: pet: PET graph
    :param: suggestions: List[PatternInfo]
    :return: List[PatternInfo]
    """
    task_suggestions = [s for s in [t for t in suggestions if type(t) == TaskParallelismInfo] if s.type is TPIType.TASK]
    for ts in task_suggestions:
        if ts.shared:
            # iterate over parent function(s)
            for parent_function in [
                pet.node_at(e[0])
                for e in pet.in_edges(ts._node.id, EdgeType.CHILD)
                if isinstance(pet.node_at(e[0]), FunctionNode)
            ]:
                # get task suggestions in parent functions scope
                for other_suggestion in task_suggestions:
                    if not (
                        line_contained_in_region(
                            other_suggestion.start_line,
                            parent_function.start_position(),
                            parent_function.end_position(),
                        )
                        and line_contained_in_region(
                            other_suggestion.end_line,
                            parent_function.start_position(),
                            parent_function.end_position(),
                        )
                    ):
                        # other suggestion not part of parent function
                        continue
                    # mark shared variables of ts as shared in other_suggestion, if not already mentioned
                    for var in ts.shared:
                        if var in other_suggestion.shared:
                            continue
                        if var in other_suggestion.private:
                            continue
                        if var in other_suggestion.first_private:
                            continue
                        other_suggestion.shared.append(var)

    return suggestions



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import pathlib
import re
import subprocess
from typing import Dict, List, Optional, Match, cast, Any

from lxml import objectify  # type: ignore


def __prune_statement(stmt_copy: str, statement: str, var_name: str, var_type: str) -> Optional[List[str]]:
    """splits a statement and performs multiple alias analyses if necessary (more than one '=' contained
    in the given statement).
    :param stmt_copy: cleaned copy of the target statement
    :param statement: statement to check for new alias definition
    :param var_name: target variable name
    :param var_type: type of the variable
    :return: List containing found aliases"""
    if stmt_copy.count(",") == 0:
        return None
    indices = [i for i, x in enumerate(stmt_copy) if x == ","]
    indices = indices + [len(stmt_copy)]
    aliases: List[str] = []
    for idx, index in enumerate(indices):
        if idx == 0:
            stmt = statement[: indices[idx]]
        elif idx == len(indices) - 1:
            stmt = statement[indices[idx - 1] + 1 : index + 1]
        else:
            stmt = statement[indices[idx - 1] + 1 : indices[idx]]
        tmp = __get_alias_from_statement(var_name, var_type, stmt)
        if tmp is not None:
            aliases += tmp
    if len(aliases) == 0:
        return None
    return aliases


def __check_obvious_pointer_type(var_name: str, rhs: str) -> bool:
    """Checks conditions for obvious pointer types.
    :param var_name: target variable name
    :param rhs: right hand side string to be analyzed
    :return: True, of None should be returned by __get_alias_statement. False, otherwise.
    """
    # var_name has index access?
    if rhs.index(var_name) + len(var_name) < len(rhs) and rhs[rhs.index(var_name) + len(var_name)] == "[":
        return True
    # '*' prior to var_name?
    if rhs.index(var_name) - 1 >= 0 and rhs[rhs.index(var_name) - 1] == "*":
        return True
    # '*' and '(' prior, no ')' prior to var_name?
    if (
        rhs.index(var_name) - 2 >= 0
        and rhs[rhs.index(var_name) - 2] == "*"
        and rhs[rhs.index(var_name) - 1] == "("
        and ")" not in rhs[rhs.index(var_name) - 2 : rhs.index(var_name) + len(var_name)]
    ):
        return True
    # '->' after var_name?
    if (
        rhs.index(var_name) + len(var_name) + 2 > len(rhs)
        and rhs[rhs.index(var_name) + len(var_name) : rhs.index(var_name) + len(var_name) + 1] == "->"
    ):
        return True
    return False


def __check_possible_pointer_type(var_name: str, rhs: str) -> bool:
    """Checks conditions for possible pointer types.
    :param var_name: target variable name
    :param rhs: right hand side string to be analyzed
    :return: True, of None should be returned by __get_alias_statement. False, otherwise.
    """
    # '&' prior to var_name?
    if rhs.index(var_name) - 1 >= 0 and not rhs[rhs.index(var_name) - 1] == "&":
        return True
    # var_name has index access?
    if rhs.index(var_name) + len(var_name) < len(rhs) and rhs[rhs.index(var_name) + len(var_name)] == "[":
        return True
    # '*' prior to '&'?
    if rhs.index(var_name) - 2 >= 0 and rhs[rhs.index(var_name) - 1] == "&" and rhs[rhs.index(var_name) - 2] == "*":
        return True
    # '*' and '(' prior, no ')' prior to '&'?
    if (
        rhs.index(var_name) - 2 >= 0
        and rhs[rhs.index(var_name) - 2] == "*"
        and rhs[rhs.index(var_name) - 1] == "("
        and ")" not in rhs[rhs.index(var_name) - 2 : rhs.index(var_name)]
    ):
        return True
    # '->' after var_name?
    if (
        rhs.index(var_name) + len(var_name) + 2 > len(rhs)
        and rhs[rhs.index(var_name) + len(var_name) : rhs.index(var_name) + len(var_name) + 1] == "->"
    ):
        return True
    return False


def __check_pointer_type(var_type: str, var_name: str, rhs: str) -> bool:
    """Distinguishes obvious and possible pointer types.
    :param var_type: type of the variable
    :param var_name: target variable name
    :param rhs: right hand side string to be analyzed
    :return: True, if None should be returned. False, otherwise."""
    # var_name is pointer type?
    if "*" in var_type:
        return __check_obvious_pointer_type(var_name, rhs)
    else:
        return __check_possible_pointer_type(var_name, rhs)


def __get_alias_from_statement(var_name: str, var_type: str, statement: str) -> Optional[List[str]]:
    """Checks if the given statement defines a new alias for var_name.
    Returns a list containing the found alias name or None, if no alias has been defined by statement.
    :param var_name: target variable name
    :param var_type: type of the variable
    :param statement: statement to check for new alias definition
    :return: None, or list containing a found alias name."""
    # prune var_name
    if ".addr" in var_name:
        var_name = var_name.replace(".addr", "")
    # prune statement
    if ":" in statement:
        statement = statement[statement.rindex(":") + 1 :]
    # var_name in statement?
    reg = r"\W" + re.escape(var_name) + r"\W"
    search_match = re.search(reg, statement)
    try:
        search_string: Optional[str] = cast(Match[str], search_match)[0]
    except TypeError:
        search_string = None
    if search_string is None:
        return None
    # operation is '=' ?
    stmt_copy = statement
    stmt_copy = stmt_copy.replace("<=>", "  ").replace("<<=", "   ").replace(">>=", "   ").replace("==", "  ")
    stmt_copy = stmt_copy.replace("!=", "  ").replace("+=", "  ").replace("-=", "  ").replace("*=", "  ")
    stmt_copy = stmt_copy.replace("/=", "  ").replace("%=", "  ").replace("<=", "  ").replace(">=", "  ")
    stmt_copy = stmt_copy.replace("&=", "  ").replace("^=", "  ").replace("|=", "  ")
    if "=" not in stmt_copy:
        return None
    # prune statement to single contained '='
    if stmt_copy.count("=") > 1:
        return __prune_statement(stmt_copy, statement, var_name, var_type)

    # var_name on left hand side?
    if stmt_copy.index(var_name) < stmt_copy.index("="):
        return None
    # var_name contained in function call?
    left_hand_side = statement[: stmt_copy.index("=")]
    right_hand_side = statement[stmt_copy.index("=") + 1 :]
    call_match = re.search(r"[\w]+(?=\().+\)", right_hand_side)
    try:
        call_string: Optional[str] = cast(Match[str], call_match)[0]
    except TypeError:
        call_string = None
    if call_string is not None:
        if var_name in call_string:
            return None
    # check pointer tye
    if __check_pointer_type(var_type, var_name, right_hand_side):
        return None
    # left hand side is single token?
    left_hand_split = left_hand_side.split(" ")
    left_hand_split = [x for x in left_hand_split if len(x) > 0]
    if len(left_hand_split) == 1:
        return left_hand_split
    # only type and modifiers on left hand side?
    if len(left_hand_split) <= 4:
        return [left_hand_split[-1]]
    return None


def __add_alias_information(
    function_information_list: List[Dict[str, Any]], statements_file: str
) -> List[Dict[str, Any]]:
    """Wrapper to gather and append alias information to the entries in function_information as a new field.
    Aliases can be found up to a depth of 2.
    Alias detection ignores scopes.
    Alias detection ignores a=b=c constellations.
    :param function_information_list: list of dictionaries representing the functions contained in cu-xml
    :param statements_file: path to file containing the output of getStatements
    :return: Input list of dictionaries, enriched with alias information."""
    result_list = []
    for function_information in function_information_list:
        file_id = function_information["id"].split(":")[0]
        body_start_line = function_information["startsAtLine"].split(":")[1]
        body_end_line = function_information["endsAtLine"].split(":")[1]
        # get such statements which occur inside current functions body
        relevant_statements = []
        with open(statements_file, "r") as sf:
            for line in sf.readlines():
                line_file_id = line[: line.index(":")]
                line_code_line = line[line.index(":") + 1 :]
                line_code_line = line_code_line[: line_code_line.index(":")]
                if file_id != line_file_id:
                    continue
                if int(body_start_line) > int(line_code_line):
                    continue
                if int(body_end_line) < int(line_code_line):
                    continue
                line = line.replace("\n", "")
                # append line + line_code_line for sorting, removed afterwards
                relevant_statements.append((line, int(line_code_line)))
        # order relevant statements by line number
        relevant_statements.sort(key=lambda x: x[1])
        relevant_statements_strings = [x[0] for x in relevant_statements]
        # get aliases for each argument
        function_information["aliases"] = []
        for arg_idx, arg_name in enumerate(function_information["args"]):
            aliases = []
            for statement in relevant_statements_strings:
                # search for first-level aliases
                statement_result = __get_alias_from_statement(
                    arg_name, function_information["arg_types"][arg_idx], statement
                )
                if statement_result is not None:
                    for state_res_entry in statement_result:
                        # first level alias found
                        # search for second level aliases
                        for inner_statement in relevant_statements_strings:
                            statement_line = statement[statement.index(":") + 1 :]
                            statement_line = statement_line[: statement_line.index(":")]
                            inner_statement_line = inner_statement[inner_statement.index(":") + 1 :]
                            inner_statement_line = inner_statement_line[: inner_statement_line.index(":")]
                            if int(inner_statement_line) < int(statement_line):
                                continue
                            inner_statement_result = __get_alias_from_statement(state_res_entry, "*", inner_statement)
                            if inner_statement_result is not None:
                                # second level alias found
                                aliases += inner_statement_result
                    aliases += statement_result
            aliases = list(set(aliases))
            function_information["aliases"].append(aliases)
        result_list.append(function_information)
    return result_list


def __get_function_information(cu_xml: str) -> List[Dict[str, Any]]:
    """Extracts information on functions from given cu_xml file and stores it in a dictionary representation.
    :param cu_xml: path to cu_xml file
    :return: List of dictionaries representing functions from cu_xml"""
    with open(cu_xml) as xml_fd:
        xml_content = ""
        for line in xml_fd.readlines():
            if not (line.rstrip().endswith("</Nodes>") or line.rstrip().endswith("<Nodes>")):
                xml_content = xml_content + line
    xml_content = "<Nodes>{0}</Nodes>".format(xml_content)
    parsed_cu = objectify.fromstring(xml_content)
    function_information = []
    for node in parsed_cu.Node:
        node.childrenNodes = str(node.childrenNodes).split(",") if node.childrenNodes else []
        if node.get("type") == "1":
            entry = dict()
            entry["name"] = node.get("name")
            entry["startsAtLine"] = node.get("startsAtLine")
            entry["endsAtLine"] = node.get("endsAtLine")
            entry["id"] = node.get("id")
            entry["args"] = []
            if hasattr(node, "funcArguments") and hasattr(node.funcArguments, "arg"):
                entry["args"] = [v.text for v in node.funcArguments.arg]
            entry["arg_types"] = []
            if hasattr(node, "funcArguments") and hasattr(node.funcArguments, "arg"):
                entry["arg_types"] = [v.get("type") for v in node.funcArguments.arg]
            function_information.append(entry)
    return function_information


def __create_statements_file(file_mapping: str, output_file: str, application_path: str):
    """Wrapper to start getStatements application for each source code file in file_mapping
    :param file_mapping: path to file_mapping file
    :param output_file: path to output file
    :param application_path: path to getStatements executable"""
    file_mapping_dict = dict()  # maps path to file-id
    with open(file_mapping) as fm:
        for line in fm.readlines():
            line = line.replace("\n", "")
            split_line = line.split("\t")
            file_mapping_dict[split_line[1]] = split_line[0]
    # execute application for each file in file_mapping
    for file in file_mapping_dict:
        process = subprocess.Popen(
            application_path + " " + file + " >> " + output_file,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        process.wait()
    # cleanup results and replace paths with file-ids
    if os.path.exists(output_file + "_tmp"):
        os.remove(output_file + "_tmp")
    with open(output_file + "_tmp", "w+") as tmp_of:
        with open(output_file, "r") as of:
            for line in of.readlines():
                line = line.replace("\n", "")
                if len(line) == 0:
                    continue
                left = line[: line.rfind(":")]
                right = line[line.rfind(":") :]
                left = re.sub(r"<.*>", "", left)
                file_path = left[: left.find(":")]
                left = left.replace(file_path, file_mapping_dict[file_path])
                line = left + right
                while " :" in line or ": " in line:
                    line = line.replace(" :", ":")
                    line = line.replace(": ", ":")
                if "=" not in line:
                    continue
                tmp_of.write(line + "\n")
    os.remove(output_file)
    os.rename(output_file + "_tmp", output_file)


def get_alias_information(file_mapping: str, cu_xml: str, temp_file: str, build_path: str) -> str:
    """Gather simple alias information for every file in filemapping and return the results in string format.
    :param file_mapping: path to filemapping-file
    :param cu_xml: path to cu_xml file
    :param temp_file: path to temporary file(s)
    :param build_path: path to discopop build directory
    :return: string, containing found aliases for each function"""
    if build_path.endswith("/"):
        build_path = build_path[:-1]

    if not os.path.isfile(file_mapping):
        raise ValueError(f'File not found: "{file_mapping}"')
    if not os.path.isfile(cu_xml):
        raise ValueError(f'File not found: "{cu_xml}"')

    # remove output file if it already exists
    if os.path.exists(temp_file + "_statements"):
        os.remove(temp_file + "_statements")
    if os.path.exists(temp_file):
        os.remove(temp_file)
    # create statements file
    __create_statements_file(
        file_mapping,
        temp_file + "_statements",
        str(pathlib.Path(build_path).joinpath("rtlib", "simple-alias-detection", "getStatements")),
    )
    # get function information file
    function_information = __get_function_information(cu_xml)
    # add alias information to function_information
    function_information = __add_alias_information(function_information, temp_file + "_statements")
    # create alias output file
    alias_str = ""
    for fn_info in function_information:
        if len(fn_info["args"]) == 0:
            continue
        for idx, alias_entry in enumerate(fn_info["aliases"]):
            if len(alias_entry) > 0:
                for alias_name in alias_entry:
                    alias_str += (
                        fn_info["id"] + ";" + fn_info["name"] + ";" + fn_info["args"][idx] + ";" + alias_name + "\n"
                    )
    # cleanup
    if os.path.exists(temp_file + "_statements"):
        os.remove(temp_file + "_statements")
    if os.path.exists(temp_file):
        os.remove(temp_file)
    return alias_str



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Tuple

from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_library.discopop_optimizer.classes.system.devices.Device import Device
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum


class GPU(Device):
    def __init__(self, frequency, thread_count, openmp_device_id, device_specific_compiler_flags):
        super().__init__(frequency, thread_count, openmp_device_id, device_specific_compiler_flags)

    def get_device_specific_pattern_info(
        self, suggestion: PatternInfo, suggestion_type: str
    ) -> Tuple[PatternInfo, str]:
        return suggestion, "gpu_" + suggestion_type

    def get_device_type(self) -> DeviceTypeEnum:
        return DeviceTypeEnum.GPU



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
from typing import Dict, List, Tuple, Set, cast

import networkx as nx  # type: ignore
from sympy import Symbol, Expr

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.CostModels.DataTransfer.DataTransferCosts import (
    add_data_transfer_costs,
)
from discopop_library.discopop_optimizer.CostModels.utilities import get_node_performance_models
from discopop_library.discopop_optimizer.DataTransfers.DataTransfers import calculate_data_transfers
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_all_function_nodes,
    get_successors,
    get_children,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


def get_locally_optimized_models(
    experiment: Experiment,
    graph: nx.DiGraph,
    substitutions: Dict[Symbol, Expr],
    environment: Experiment,
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution],
) -> Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]]:
    result_dict: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]] = dict()
    all_function_ids = get_all_function_nodes(graph)
    all_function_nodes: List[FunctionRoot] = [cast(FunctionRoot, data_at(graph, fn_id)) for fn_id in all_function_ids]
    for function_node in all_function_ids:
        # get a list of all decisions that have to be made
        decisions_to_be_made = __find_decisions(graph, function_node)
        locally_optimal_choices: List[int] = []

        # create performance models for individual options of decisions
        for decision_options in decisions_to_be_made:
            decision_models: List[Tuple[int, Tuple[CostModel, ContextObject]]] = []
            for decision in decision_options:
                try:
                    # create a performance model for the specific decision
                    performance_models = get_node_performance_models(
                        experiment,
                        graph,
                        function_node,
                        set(),
                        all_function_nodes,
                        restrict_to_decisions={decision},
                        do_not_allow_decisions=set([o for o in decision_options if o != decision]),
                        ignore_node_costs=[
                            cast(FunctionRoot, data_at(graph, function_node)).node_id
                        ],  # ignore first node to prevent duplication of function costs
                    )
                    # calculate and append necessary data transfers to the models
                    performance_models_with_transfers = calculate_data_transfers(
                        graph, {cast(FunctionRoot, data_at(graph, function_node)): performance_models}, experiment
                    )

                    # calculate and append costs of data transfers to the performance models
                    complete_performance_models = add_data_transfer_costs(
                        graph, performance_models_with_transfers, environment
                    )
                    # add performance models to decision models for the later selection of the best candidate
                    for function in complete_performance_models:
                        for pair in complete_performance_models[function]:
                            decision_models.append((decision, pair))
                except ValueError as ex:
                    print(ex)
                    print("==> Ignoring Decision: ", decision)
                    continue

            # iteratively apply variable substitutions
            decision_models_with_substitutions: List[Tuple[int, Tuple[CostModel, ContextObject, CostModel]]] = []
            # initialize substitution
            for decision, pair in decision_models:
                model, context = pair
                decision_models_with_substitutions.append((decision, (model, context, copy.deepcopy(model))))

            modification_found = True
            while modification_found:
                modification_found = False
                for decision, tpl in decision_models_with_substitutions:
                    model, context, substituted_model = tpl

                    # apply substitutions to parallelizable costs
                    tmp_model = substituted_model.parallelizable_costs.subs(substitutions)
                    if tmp_model != substituted_model.parallelizable_costs:
                        modification_found = True
                    substituted_model.parallelizable_costs = substituted_model.parallelizable_costs.subs(substitutions)

                    # apply substitutions to sequential costs
                    tmp_model = substituted_model.sequential_costs.subs(substitutions)
                    if tmp_model != substituted_model.sequential_costs:
                        modification_found = True
                    substituted_model.sequential_costs = substituted_model.sequential_costs.subs(substitutions)

            #                    decision_models_with_substitutions.append(
            ##                        (decision, (model, context, substituted_model))
            #                   )

            # set free symbol ranges and distributions for comparisons
            for decision, tpl in decision_models_with_substitutions:
                model, context, substituted_model = tpl
                model.free_symbol_ranges = free_symbol_ranges
                model.free_symbol_distributions = free_symbol_distributions
                substituted_model.free_symbol_ranges = free_symbol_ranges
                substituted_model.free_symbol_distributions = free_symbol_distributions

            # find minimum in decision_models
            unpacked_models: List[Tuple[int, CostModel, CostModel]] = []
            for decision, tpl in decision_models_with_substitutions:
                model, context, substituted_model = tpl
                unpacked_models.append((decision, model, substituted_model))
            if len(unpacked_models) == 0:
                continue
            minimum = sorted(unpacked_models, key=lambda x: x[2])[0]
            locally_optimal_choices.append(minimum[0])

        if len(locally_optimal_choices) == 0:
            continue

        # construct locally optimal model
        performance_models = get_node_performance_models(
            experiment,
            graph,
            function_node,
            set(),
            all_function_nodes,
            restrict_to_decisions=set(locally_optimal_choices),
            ignore_node_costs=[
                cast(FunctionRoot, data_at(graph, function_node)).node_id
            ],  # ignore first node to prevent duplication of function costs
        )
        # calculate and append necessary data transfers to the models
        performance_models_with_transfers = calculate_data_transfers(
            graph, {cast(FunctionRoot, data_at(graph, function_node)): performance_models}, experiment
        )

        # calculate and append costs of data transfers to the performance models
        complete_performance_models = add_data_transfer_costs(graph, performance_models_with_transfers, environment)
        # merge dictionaries
        result_dict = {**result_dict, **complete_performance_models}

    return result_dict


def __find_decisions(graph: nx.DiGraph, root_node: int) -> Set[Tuple[int, ...]]:
    result_set: Set[Tuple[int, ...]] = set()
    successors = get_successors(graph, root_node)
    children = get_children(graph, root_node)

    # check if children have decisions
    for child in children:
        result_set.update(__find_decisions(graph, child))

    # check if current node has a decision
    if len(successors) > 1:
        local_decision = list()
        for successor in successors:
            local_decision.append(successor)
        result_set.add(tuple(local_decision))

    # check if successors have decisions
    for successor in successors:
        result_set.update(__find_decisions(graph, successor))

    return result_set
